from typing import List
from portpy.photon.beam import Beams
from portpy.photon.structures import Structures
import os
from portpy.photon.clinical_criteria import ClinicalCriteria
from portpy.photon.influence_matrix import InfluenceMatrix
from portpy.photon.visualization import Visualization
from portpy.photon.optimization import Optimization
from portpy.photon.utils import *


class Plan:
    """
    A class representing a plan

    - **Attributes** ::

        :param opt_beamlets_PTV_margin_mm: For each beam, we often only include the beamlets that are within
            few millimetres of the projection of the PTV (tumor) into that beam. It is because the beamlets
            that are far from the PTV projection mainly deliver radiation to the healthy tissues not PTV. Default is 3mm
        :type opt_beamlets_PTV_margin_mm: int
        :param beams_dict: an object of class Beams that contains information about the beams_dict used in the treatment plan.
        :type beams: object
        :param structures: an object of class Structures that contains information about the structures present in the patient's CT scan.
        :type structures: object
        :param inf_matrix: object of class Influence matrix
        :type inf_matrix: object
        :param clinial_criteria: an object of class ClincialCriteria that contains information about the goals and constraints of the treatment plan
        :type inf_matrix: object
        :param ct: dictionary containing ct metadata. It includes metadata about important ct parameters like resolution, ct in HU etc.
        :type ct: dict

    - **Methods** ::

        :create_inf_matrix(beamlet_width_mm, beamlet_height_mm,
                          down_sample_xyz):
            Create object of Influence Matrix class
        :get_prescription():
            Return prescription for the plan
        :get_num_of_fractions()
            Return number of fractions for the plan




    """

    def __init__(self, patient_id: str, data_dir: str = None, beam_ids: List[int] = None,
                 opt_beamlets_PTV_margin_mm: int = 3, load_inf_matrix_full: bool = False,
                 protocol_name='Lung_Default_2Gy_30Fr', protocol_type='Default') -> None:
        """
        Creates an object of Plan class for the specified patient

        :param patient_id: the patient that we're creating a plan for (name of the folder)
        :param data_dir: the name of the folder that includes all the patients' data.
            If None, then ''current-folder\Data'' is used
        :param beam_ids: the indices of the beams_dict used for creating the plan object.
            If None, the beams_dict selected by an expert (human) physicist for the specified patient would be used
        :param opt_beamlets_PTV_margin_mm: For each beam, we often only include the beamlets that are within
            few millimetres of the projection of the PTV (tumor) into that beam. It is because the beamlets
            that are far from the PTV projection mainly deliver radiation to the healthy tissues not PTV. Default is 3mm
        :param load_inf_matrix_full: If set to true, it will load full influence matrix from the data

        :Example:

        >>> my_plan = Plan(patient_id = r"Lung_Patient_1", path = r"c:\Data", beam_ids = [0,1,2,3,4,5,6], opt_beamlets_PTV_margin_mm=3)
        """

        if data_dir is None:
            data_dir = os.path.join('../..', 'data')
        patient_folder_path = os.path.join(data_dir, patient_id)

        # read all the meta data for the specified patient and protocol
        meta_data = load_metadata(patient_folder_path)

        # filter metadata for the given beam_indices.
        # The meta_data originally includes all the beams_dict.  get_plan_beams only keeps the requested beams_dict
        meta_data = self.get_plan_beams(beam_ids=beam_ids, meta_data=meta_data)
        # Load all the data (e.g., influence matrix, voxel coordinates).
        # meta_data does not include the large numeric data stored in .h5 files
        data = load_data(meta_data=meta_data, pat_dir=patient_folder_path, load_inf_matrix_full=load_inf_matrix_full)
        self.opt_beamlets_PTV_margin_mm = opt_beamlets_PTV_margin_mm
        self.beams = Beams(data['beams_dict'])  # create beams_dict object
        self.structures = Structures(data['structures'], data['opt_voxels'])  # create structures object
        self.ct = data['ct']  # create ct attribute containing ct information as dictionary
        self.patient_id = patient_id
        self.clinical_criteria = ClinicalCriteria(data['clinical_criteria'])  # create clinical criteria object
        is_full = False
        if load_inf_matrix_full:  # check if full influence matrix is requested
            is_full = True

        self.inf_matrix = InfluenceMatrix(self, is_full=is_full)  # create influence matrix object

    @staticmethod
    def get_plan_beams(beam_ids: List[int] = None, meta_data: dict = None) -> dict:
        """
        Create and return a copy of meta_data with only including the requested beams_dict (beam_ids)


        :param beam_ids: the indices of the beams_dict to be included. If None, the planner's beams_dict are used
        :param meta_data: the dictionary including all the beams_dict
        :return: returns the meta_data dictionary only including the requested beams_dict in format of:
            dict: {
                   'structures': {'name': list(str), 'volume_cc': list(float), }
                   'opt_voxels': {'_ct_voxel_resolution_xyz_mm': list(float),}
                  }
        """
        if beam_ids is None:  # if beam_ids not included, then the beams_dict
            # selected by an expert human planner would be used
            beam_ids = meta_data['planner_beam_ids']['IDs']
        meta_data_req = meta_data.copy()
        del meta_data_req['beams_dict']  # remove previous beams_dict
        beamReq = dict()
        for i in range(len(beam_ids)):
            if beam_ids[i] in meta_data['beams_dict']['ID']:
                ind = meta_data['beams_dict']['ID'].index(beam_ids[i])
                for key in meta_data['beams_dict']:
                    beamReq.setdefault(key, []).append(meta_data['beams_dict'][key][ind])
            else:
                print('beam id {} is not available'.format(beam_ids[i]))
        meta_data_req['beams_dict'] = beamReq
        return meta_data_req

    def save_plan(self, plan_name: str = None, path: str = None) -> None:
        """

        Save pickled file for plan object

        :param plan_name: create the name of the pickled file of plan object. If none, it will save with the name as 'my_plan'
        :param path: if path is set, plan object will be pickled and saved in path directory else it will save current project directory
        :return: save pickled object of class Plan

        :Example:
        >>> my_plan.save_plan(plan_name='my_plan', path=r"path/to/save_plan")
        """
        save_plan(self, plan_name=plan_name, path=path)

    @staticmethod
    def load_plan(plan_name: str = None, path: str = None):
        """
        Load pickle file of the plan object.

        :param plan_name: plan_name of the object of class Plan. It None, it will try to look for plan name called 'my_plan'
        :param path: if path is set, plan object will be load from path directory else current project directory
        :return: load pickled object of class Plan

        :Example:
        >>> Plan.load_plan(plan_name='my_plan', path=r"path/for/loading_plan")
        """

        return load_plan(plan_name=plan_name, path=path)

    @staticmethod
    def load_optimal_sol(sol_name: str, path: str = None) -> dict:
        """

        Load optimal solution dictionary got from optimization

        :param sol_name: name of the optimal solution to be loaded.
        :param path: if path is set, plan object will be load from path directory else current directory
        :return: load solution

        :Example:
        >>> sol = Plan.load_optimal_sol(sol_name='sol', path=r'path/for/loading_sol')
        """
        return load_optimal_sol(sol_name=sol_name, path=path)

    @staticmethod
    def save_optimal_sol(sol: dict, sol_name: str, path: str = None) -> None:
        """
        Save the optimal solution dictionary from optimization

        :param sol: optimal solution dictionary
        :param sol_name: name of the optimal solution saved
        :param path: if path is set, plan object will be load from path directory else current directory
        :return: save pickled file of optimal solution dictionary

        :Example:
        >>> my_plan.save_optimal_sol(sol=sol, sol_name='sol', path=r'path/to/save_solution')
        """
        save_optimal_sol(sol=sol, sol_name=sol_name, path=path)

    def create_inf_matrix(self, beamlet_width_mm: float = 2.5, beamlet_height_mm: float = 2.5,
                          opt_vox_xyz_res_mm: List[float] = None,
                          structure: str = 'PTV', is_full: bool = False) -> InfluenceMatrix:
        """
                Create a influence matrix object for Influence Matrix class

                :param is_full: Defaults to False. If True, it will create full influence matrix
                :param beamlet_width_mm: beamlet width in mm. It should be multiple of 2.5, defaults to 2.5
                :param beamlet_height_mm: beamlet height in mm. It should be multiple of 2.5, defaults to 2.5
                :param structure: target structure for creating BEV beamlets, defaults to 'PTV'
                :param opt_vox_xyz_res_mm: It down-samples optimization voxels as factor of ct resolution
                    e.g. opt_vox_xyz_res = [5*ct.res.x,5*ct.res.y,1*ct.res.z]. It will down-sample optimization voxels with 5 * ct res. in x direction, 5 * ct res. in y direction and 1*ct res. in z direction.
                    defaults to None. When None it will use the original optimization voxel resolution.
                :returns: object of influence Matrix class

                :Example:
                >>> inf_matrix = my_plan.create_inf_matrix(beamlet_width_mm=5, beamlet_height_mm=5, opt_vox_xyz_res_mm=[5,5,1], structure=structure)
                """
        return InfluenceMatrix(self, beamlet_width_mm=beamlet_width_mm, beamlet_height_mm=beamlet_height_mm,
                               opt_vox_xyz_res_mm=opt_vox_xyz_res_mm, is_full=is_full)

    def get_prescription(self) -> float:
        """

        :return: prescription in Gy
        """
        pres = self.clinical_criteria.clinical_criteria_dict['pres_per_fraction_gy'] * \
               self.clinical_criteria.clinical_criteria_dict[
                   'num_of_fractions']
        return pres

    def get_num_of_fractions(self) -> float:
        """

        :return: number of fractions to be delivered
        """
        return self.clinical_criteria.clinical_criteria_dict['num_of_fractions']

    def get_disease_site(self) -> float:
        """

        :return: number of fractions to be delivered
        """
        return self.clinical_criteria.clinical_criteria_dict['disease_site']

    def get_ct_res_xyz_mm(self) -> List[float]:
        """

        :return: number of fractions to be delivered
        """
        return self.ct['resolution_xyz_mm']

    def plot_dvh(self, sol, dose_1d=None, structs=None, dose_scale: Visualization.dose_type = "Absolute(Gy)",
                 volume_scale: Visualization.volume_type = "Relative(%)", **options):
        """
                Create dvh plot for the selected structures

                :param sol: optimal sol dictionary
                :param dose_1d: dose_1d in 1d voxels
                :param structs: structures to be included in dvh plot
                :param volume_scale: volume scale on y-axis. Default= Absolute(cc). e.g. volume_scale = "Absolute(cc)" or volume_scale = "Relative(%)"
                :param dose_scale: dose_1d scale on x axis. Default= Absolute(Gy). e.g. dose_scale = "Absolute(Gy)" or dose_scale = "Relative(%)"
                :keyword style (str): line style for dvh curve. default "solid". can be "dotted", "dash-dotted".
                :keyword width (int): width of line. Default 2
                :keyword colors(list): list of colors
                :keyword legend_font_size: Set legend_font_size. default 10
                :keyword figsize: Set figure size for the plot. Default figure size (12,8)
                :keyword create_fig: Create a new figure. Default True. If False, append to the previous figure
                :keyword title: Title for the figure
                :keyword filename: Name of the file to save the figure in current directory
                :keyword show: Show the figure. Default is True. If false, next plot can be append to it
                :keyword norm_flag: Use to normalize the plan. Default is False.
                :keyword norm_volume: Use to set normalization volume. default is 90 percentile.
                :return: dvh plot for the selected structures
                """

        Visualization.plot_dvh(self, sol=sol, dose_1d=dose_1d, structs=structs, dose_scale=dose_scale,
                               volume_scale=volume_scale, **options)

    def run_IMRT_fluence_map_CVXPy(self, inf_matrix: InfluenceMatrix = None, solver='MOSEK'):
        Optimization.run_IMRT_fluence_map_CVXPy(self, inf_matrix=inf_matrix, solver=solver)

    def add_rinds(self, rind_params: List[dict], inf_matrix=None):
        """
        Example for
        rind_params = [{'rind_name': 'RIND_0', 'ref_structure': 'PTV, 'margin_start_mm': 2, 'margin_end_mm': 10, 'max_dose_gy': 10}]

        :param rind_params: rind_params as dictionary
        :param inf_matrix: object of class inf_matrix
        :return: save rinds to plan object
        """

        if inf_matrix is None:
            inf_matrix = self.inf_matrix
        print('creating rinds..')

        ct_to_dose_map = inf_matrix.opt_voxels_dict['ct_to_dose_voxel_map'][0]
        dose_mask = ct_to_dose_map >= 0
        dose_mask = dose_mask.astype(int)
        self.structures.create_structure('dose_mask', dose_mask)

        for ind, param in enumerate(rind_params):
            rind_name = param['rind_name']
            first_dummy_name = '{}_{}'.format(param['ref_structure'], param['margin_start_mm'])
            second_dummy_name = '{}_{}'.format(param['ref_structure'], param['margin_end_mm'])
            self.structures.expand(param['ref_structure'], margin_mm=param['margin_start_mm'],
                                   new_structure=first_dummy_name)
            if param['margin_end_mm'] == 'inf':
                param['margin_end_mm'] = 500
            self.structures.expand(param['ref_structure'], margin_mm=param['margin_end_mm'],
                                   new_structure=second_dummy_name)
            self.structures.subtract(second_dummy_name, first_dummy_name, str1_sub_str2=rind_name)
            self.structures.delete_structure(first_dummy_name)
            self.structures.delete_structure(second_dummy_name)
            self.structures.intersect(rind_name, 'dose_mask', str1_and_str2=rind_name)
        self.structures.delete_structure('dose_mask')

        print('rinds created!!')

        for param in rind_params:
            inf_matrix.set_opt_voxel_idx(self, structure_name=param['rind_name'])
            # add rind constraint
            parameters = {'structure_name': param['rind_name']}
            # total_pres = self.get_prescription()
            if 'max_dose_gy' in param:
                constraints = {'limit_dose_gy': param['max_dose_gy']}
                self.clinical_criteria.add_criterion(criterion='max_dose', parameters=parameters,
                                                     constraints=constraints)
            if 'mean_dose_gy' in param:
                constraints = {'limit_dose_gy': param['mean_dose_gy']}
                self.clinical_criteria.add_criterion(criterion='mean_dose', parameters=parameters,
                                                     constraints=constraints)

    @staticmethod
    def plot_fluence_2d(beam_id: int, sol: dict = None, **options):
        """
        plot fluence in 2d for beam_id
        :param beam_id: beam_id of the beam
        :param sol: solution dictionary after optimization
        :return: 2d optimal fluence plot
        """
        return Visualization.plot_fluence_2d(beam_id=beam_id, sol=sol, **options)

    @staticmethod
    def plot_fluence_3d(beam_id: int, sol: dict = None, **options):
        """
        plot fluence in 3d for beam_id
        :param sol: solution after optimization
        :param beam_id: beam_id of the beam
        :return: 3d optimal fluence plot
        """
        return Visualization.plot_fluence_3d(beam_id=beam_id, sol=sol, **options)

    def save_nrrd(self, sol: dict, data_dir: str = None):
        """
        save nrrd in the path directory else save in patient data directory
        :param sol: optimal solution dict
        :param data_dir: save nrrd images of ct, dose_1d and structure set in path directory
        :return: save nrrd images in path
        """
        save_nrrd(self, sol=sol, data_dir=data_dir)

    def view_in_slicer(self, slicer_path=None, img_dir=None):
        """
        view ct, dose_1d and structures in slicer
        :param slicer_path:
        :param img_dir:
        :return:
        """
        Visualization.view_in_slicer(self, slicer_path=slicer_path, data_dir=img_dir)



from scipy import interpolate
import numpy as np


class Evaluation:

    @staticmethod
    def get_dose(sol: dict, struct: str, volume_per: float, dose_1d: np.ndarray = None, weight_flag: bool = True) -> float:
        """
        Get dose_1d at volume percentage

        :param sol: solution dictionary
        :param dose_1d: dose_1d in 1d
        :param struct: structure name for which to get the dose_1d
        :param volume_per: query the dose_1d at percentage volume
        :param weight_flag: for non uniform voxels weight flag always True
        :return: dose_1d at volume_percentage

        :Example:

        >>> Evaluation.get_dose(sol=sol, struct='PTV', volume_per=90)

        """
        x, y = Evaluation.get_dvh(sol, dose_1d=dose_1d, struct=struct, weight_flag=weight_flag)
        f = interpolate.interp1d(100 * y, x)

        return f(volume_per)

    @staticmethod
    def get_volume(sol: dict, struct: str, dose_value_gy: float, dose_1d: np.ndarray = None, weight_flag: bool = True) -> float:
        """
        Get volume at dose_1d value in Gy

        :param sol: solution dictionary
        :param dose_1d: dose_1d in 1d
        :param struct: structure name for which to get the dose_1d
        :param dose_value_gy: query the volume at dose_value
        :param weight_flag: for non uniform voxels weight flag always True
        :return: dose_1d at volume_percentage

        :Example:

        >>> Evaluation.get_volume(sol=sol, struct='PTV', dose_value_gy=60)

        """
        x, y = Evaluation.get_dvh(sol, dose_1d=dose_1d, struct=struct, weight_flag=weight_flag)
        x1, indices = np.unique(x, return_index=True)
        y1 = y[indices]
        f = interpolate.interp1d(x1, 100 * y1)
        if dose_value_gy > max(x1):
            print('Warning: dose_1d value {} is greater than max dose_1d for {}'.format(dose_value_gy, struct))
            return 0
        else:
            return f(dose_value_gy)

    @staticmethod
    def get_dvh(sol: dict, struct: str, dose_1d: np.ndarray = None, weight_flag: bool = True):
        """
        Get dvh for the structure

        :param sol: optimal solution dictionary
        :param dose_1d: dose_1d which is not in solution dictionary
        :param struct: structure name
        :param weight_flag: for non uniform voxels weight flag always True
        :return: x, y --> dvh for the structure

        :Example:

        >>> Evaluation.get_dvh(sol=sol, struct='PTV')

        """
        inf_matrix = sol['inf_matrix']
        vox = inf_matrix.get_opt_voxels_idx(struct)
        if dose_1d is None:
            dose_1d = sol['dose_1d']
        org_sort_dose = np.sort(dose_1d[vox])
        sort_ind = np.argsort(dose_1d[vox])
        org_sort_dose = np.append(org_sort_dose, org_sort_dose[-1] + 0.01)
        x = org_sort_dose
        if weight_flag:
            # org_points_sort_spacing = my_plan._structures.opt_voxels_dict['dose_voxel_resolution_XYZ_mm']
            # org_points_sort_volume = org_points_sort_spacing[:, 0] * org_points_sort_spacing[:,
            #                                                          1] * org_points_sort_spacing[:, 2]
            # sum_weight = np.sum(org_points_sort_volume)
            org_weights = inf_matrix.get_opt_voxels_size(struct)
            org_sort_weights = org_weights[sort_ind]
            sum_weight = np.sum(org_sort_weights)
            y = [1]
            for j in range(len(org_sort_weights)):
                y.append(y[-1] - org_sort_weights[j] / sum_weight)
        else:
            y = np.ones(len(vox) + 1) - np.arange(0, len(vox) + 1) / len(vox)
        y[-1] = 0
        y = np.array(y)
        return x, y

    @staticmethod
    def get_max_dose(sol: dict, struct: str, dose_1d=None) -> float:
        """
        Get maximum dose_1d for the structure

        :param sol: optimal solution dictionary
        :param dose_1d: dose_1d which is not in solution dictionary
        :param struct: structure name

        :return: maximum dose_1d for the structure
        """
        inf_matrix = sol['inf_matrix']
        vox = inf_matrix.get_opt_voxels_idx(struct)
        if dose_1d is None:
            dose_1d = sol['dose_1d']
        return np.max(dose_1d[vox])

    @staticmethod
    def get_mean_dose(sol: dict, struct: str, dose_1d=None) -> np.ndarray:
        """
                Get mean dose_1d for the structure

                :param sol: optimal solution dictionary
                :param dose_1d: dose_1d which is not in solution dictionary
                :param struct: structure name

                :return: mean dose_1d for the structure
                """
        inf_matrix = sol['inf_matrix']
        vox = inf_matrix.get_opt_voxels_idx(struct)
        if dose_1d is None:
            dose_1d = sol['dose_1d']
        return np.mean(dose_1d[vox])





import numpy as np
from copy import deepcopy


class Structures:
    """
    A class representing structures.

    - **Attributes** ::

        :param structures_dict: structure dictionary that contains information about the structures present in the patient's CT scan.
        :type structures_dict: dict
        :param opt_voxels_dict: It contains information about optimization voxels in the form of dictionary
        :type opt_voxels_dict: dict


    - **Methods** ::

        :get_volume_cc(struct):
            Get volume in cc for the structure
        :expand(structure, margin_mm, new_structure):
            Expand the structure with margin_mm in mm
        :shrink(structure, margin_mm, new_structure):
            Shrink the structure with margin_mm in mm
        :union(str_1, str_2, str1_or_str2):
            Create union of two structures str_1 and str_2
        :intersect(str_1, str_2, str1_and_str2):
            Create intersect of two structures str_1 and str_2


    """

    def __init__(self, structures: dict, opt_voxels: dict) -> None:
        """

        :param structures: structures dictionary that contains information about the structures present in the patient's CT scan.
        :param opt_voxels: optimization voxels dictionary containing data about optimization voxels
        """
        self.structures_dict = structures
        self.opt_voxels_dict = opt_voxels
        self.opt_voxels_dict['name'] = structures['name']
        self._ct_voxel_resolution_xyz_mm = deepcopy(self.opt_voxels_dict['ct_voxel_resolution_xyz_mm'])
        self.preprocess_structures()

    def get_structures(self) -> list:
        """
        Returns all the structure names as list
        :return:
        """
        return self.structures_dict['name']

    def get_volume_cc(self, structure_name: str):
        """
        Get volume in cc for the structure

         :param structure_name: name of the structure in plan
         :return: volume of the structure
         """
        ind = self.structures_dict['name'].index(structure_name)
        return self.structures_dict['volume_cc'][ind]

    def get_fraction_of_vol_in_calc_box(self, structure_name: str):
        """
        Get fraction of volume in calc box for the structure

         :param structure_name: name of the structure in plan
         :return: volume of the structure
         """
        ind = self.structures_dict['name'].index(structure_name)
        return self.structures_dict['fraction_of_vol_in_calc_box'][ind]

    def preprocess_structures(self):
        """
        preprocess structures to create optimization voxel indices for the structure
        :return:
        """
        self.opt_voxels_dict['voxel_idx'] = [None] * len(self.structures_dict['name'])
        self.opt_voxels_dict['voxel_size'] = [None] * len(self.structures_dict['name'])
        for i in range(len(self.structures_dict['name'])):

            vox_3d = self.structures_dict['structure_mask_3d'][i] * self.opt_voxels_dict['ct_to_dose_voxel_map'][0]
            # my_plan.structures_dict['voxel_idx'][i] = np.unique(vox_3d[vox_3d > 0])
            vox, counts = np.unique(vox_3d[vox_3d > 0], return_counts=True)
            self.opt_voxels_dict['voxel_idx'][i] = vox
            self.opt_voxels_dict['voxel_size'][i] = counts * np.prod(self._ct_voxel_resolution_xyz_mm)  # calculate weight for each voxel
            # self.opt_voxels_dict['voxel_size'][i] = counts / np.max(counts)  # calculate weight for each voxel

    def create_structure(self, new_structure: str, mask_3d: np.ndarray) -> None:
        """
        Create a new structure and append its mask to the structures_dict

        :param new_structure: name of the new structure
        :param mask_3d: 3d mask for the structure
        :return: create new_structure
        """
        for key in self.structures_dict.keys():
            if key == 'name':
                self.structures_dict['name'].append(new_structure)
            elif key == 'structure_mask_3d':
                self.structures_dict['structure_mask_3d'].append(mask_3d)
            else:
                self.structures_dict[key].append(None)

    def modify_structure(self, structure: str, mask_3d: np.ndarray):
        """
        :param structure: name of the structure to be modified
        :param mask_3d: 3d mask for the structure
        :return: modify structure
        """
        ind = self.structures_dict['name'].index(structure)
        for key in self.structures_dict.keys():
            if key == 'name':
                self.structures_dict['name'][ind] = structure
            elif key == 'structure_mask_3d':
                self.structures_dict['structure_mask_3d'][ind] = mask_3d
            else:
                self.structures_dict[key][ind] = None

    def delete_structure(self, structure: str):
        """
        :param structure: structure to be removed
        :return:
        """
        ind = self.structures_dict['name'].index(structure)
        for key in self.structures_dict.keys():
            del self.structures_dict[key][ind]

    def union(self, str_1: str, str_2: str, str1_or_str2: str) -> None:
        """
        Create union of two structures str_1 and str_2. If str1_or_str2 is not in structures dict,
        it will create new structures. If str1_or_str2 is in structure_dict, it will modify the structure

        :param str_1: structure name for the 1st structure
        :param str_2: structure name for the 2nd structure
        :param str1_or_str2: structure name for the union of  structure 1 and 2
        :return: create union of the structures and save it to structures list
        """
        if str1_or_str2 is None:
            raise Exception("str1_or_str2 need to be provided")
        ind1 = self.structures_dict['name'].index(str_1)
        ind2 = self.structures_dict['name'].index(str_2)
        mask_3d_1 = self.structures_dict['structure_mask_3d'][ind1]
        mask_3d_2 = self.structures_dict['structure_mask_3d'][ind2]
        new_mask_3d_1 = mask_3d_1 | mask_3d_2
        if str1_or_str2 not in self.structures_dict['name']:
            self.create_structure(new_structure=str1_or_str2, mask_3d=new_mask_3d_1)
        else:
            self.modify_structure(structure=str1_or_str2, mask_3d=new_mask_3d_1)

    def intersect(self, str_1: str, str_2: str, str1_and_str2: str) -> None:
        """
        Create intersection of two structures str_1 and str_2. If str1_and_str2 is not in structures dict,
        it will create new structures. If str1_and_str2 is in structure_dict, it will modify the structure

        :param str_1: structure name for the 1st structure
        :param str_2: structure name for the 2nd structure
        :param str1_and_str2: structure name for the intersection of  structure 1 and 2
        :return: create intersection of the structures and save it to structures list
        """
        if str1_and_str2 is None:
            raise Exception("str1_and_str2 need to be provided")
        ind1 = self.structures_dict['name'].index(str_1)
        ind2 = self.structures_dict['name'].index(str_2)
        mask_3d_1 = self.structures_dict['structure_mask_3d'][ind1]
        mask_3d_2 = self.structures_dict['structure_mask_3d'][ind2]
        new_mask_3d_1 = mask_3d_1 & mask_3d_2
        if str1_and_str2 not in self.structures_dict['name']:
            self.create_structure(new_structure=str1_and_str2, mask_3d=new_mask_3d_1)
        else:
            self.modify_structure(structure=str1_and_str2, mask_3d=new_mask_3d_1)

    def subtract(self, str_1: str, str_2: str, str1_sub_str2: str) -> None:
        """
        :param str_1: structure name for the 1st structure
        :param str_2: structure name for the 2nd structure
        :param str1_sub_str2: structure name for subtracting 2 from 1
        :return: create structure1 - structure2 and save it to list
        """
        if str1_sub_str2 is None:
            raise Exception("str1_sub_str2 need to be provided")
        ind1 = self.structures_dict['name'].index(str_1)
        ind2 = self.structures_dict['name'].index(str_2)
        mask_3d_1 = self.structures_dict['structure_mask_3d'][ind1]
        mask_3d_2 = self.structures_dict['structure_mask_3d'][ind2]
        new_mask_3d_1 = mask_3d_1 - mask_3d_2
        new_mask_3d_1[new_mask_3d_1 < 0] = np.uint8(0)
        if str1_sub_str2 not in self.structures_dict['name']:
            self.create_structure(new_structure=str1_sub_str2, mask_3d=new_mask_3d_1)
        else:
            self.modify_structure(structure=str1_sub_str2, mask_3d=new_mask_3d_1)

    def expand(self, structure: str, margin_mm: float, new_structure: str) -> None:
        """

        Expand the structure with the given margin_mm.

        :param structure: structure name to expand
        :param margin_mm: margin_mm in mm
        :param new_structure: new structure name. if same as structure name,
        expand the same structure. else create new structure
        :return: expand and save the structure mask in structures dictionary
        """
        # from skimage import data, morphology, transform
        from scipy import ndimage
        ind = self.structures_dict['name'].index(structure)
        mask_3d = self.structures_dict['structure_mask_3d'][ind]

        # getting kernel size for expansion or shrinking
        if margin_mm > 0:
            num_voxels = np.round(margin_mm / np.asarray(self._ct_voxel_resolution_xyz_mm)).astype(int)

            kernel = ndimage.generate_binary_structure(rank=3, connectivity=2)  # ball kernel
            # creating different iterations along z and xy directions
            num_iterations = int(num_voxels[0] / num_voxels[2])
            iterations_in_one_step = int(np.round(num_voxels[0] / num_iterations))
            # margin_mask_3d = ndimage.binary_dilation(mask_3d, structure=struct).astype(mask_3d.dtype)
            for i in range(num_iterations):
                if i == 0:
                    margin_mask_3d = ndimage.binary_dilation(mask_3d, structure=kernel,
                                                             iterations=iterations_in_one_step).astype(mask_3d.dtype)
                else:
                    flat = np.copy(kernel)
                    flat[0, :, :] = 0
                    flat[-1, :, :] = 0
                    margin_mask_3d = ndimage.binary_dilation(margin_mask_3d, structure=flat,
                                                             iterations=iterations_in_one_step).astype(mask_3d.dtype)
        elif margin_mm == 0:
            margin_mask_3d = mask_3d
        else:
            raise ValueError('Invalid margin {}'.format(margin_mm))
        if new_structure not in self.structures_dict['name']:
            self.create_structure(new_structure=new_structure, mask_3d=margin_mask_3d)
        else:
            self.modify_structure(structure=structure, mask_3d=margin_mask_3d)

    def shrink(self, structure: str, margin_mm: float, new_structure: str) -> None:
        """
        Shrink the structure with margin_mm.

            :param structure: structure name to shrink
            :param margin_mm: margin_mm in mm
            :param new_structure: new structure name. if same as structure, shrink the same structure. else create new structure
            :return: shrink and save the structure in structures_dictionary

        """

        from scipy import ndimage
        ind = self.structures_dict['name'].index(structure)
        mask_3d = self.structures_dict['structure_mask_3d'][ind]

        # getting kernel size for expansion or shrinking
        num_voxels = np.round(margin_mm / np.asarray(self._ct_voxel_resolution_xyz_mm)).astype(int)
        struct = ndimage.generate_binary_structure(rank=3, connectivity=2)  # ball kernel
        # creating different iterations along z and xy directions
        num_iterations = int(num_voxels[0] / num_voxels[2])
        iterations_in_one_step = int(np.round(num_voxels[0] / num_iterations))
        for i in range(num_iterations):
            if i == 0:
                margin_mask_3d = ndimage.binary_erosion(mask_3d, structure=struct,
                                                        iterations=iterations_in_one_step).astype(mask_3d.dtype)
            else:
                flat = np.copy(struct)
                flat[0, :, :] = 0
                flat[-1, :, :] = 0
                margin_mask_3d = ndimage.binary_erosion(margin_mask_3d, structure=flat,
                                                        iterations=iterations_in_one_step).astype(mask_3d.dtype)

            # margin_mask_3d = ndimage.binary_dilation(mask_3d, structure=structure).astype(mask_3d.dtype)
        if new_structure not in self.structures_dict['name']:
            self.create_structure(new_structure=new_structure, mask_3d=margin_mask_3d)
        else:
            self.modify_structure(structure=structure, mask_3d=margin_mask_3d)



from .plan import Plan
from .optimization import Optimization as Optimize
from .visualization import Visualization as Visualize
from .evaluation import Evaluation as Evaluate
from portpy.photon.utils import *
from .cvxpy_prob import CvxPyProb



from portpy.photon.influence_matrix import InfluenceMatrix


def sol_change_inf_matrix(sol: dict, inf_matrix: InfluenceMatrix) -> dict:
    """
    Create a new solution by changing the basis of current solution.
    It will create a solution with same number of beamlets and voxels as inf_matrix


    :param sol: solution for which influence matrix is changed
    :param inf_matrix: object of class Influence matrix
    :return: new solution dictionary having same number of beamlets and voxels as inf_matrix
    """
    new_sol = dict()
    if sol['optimal_intensity'].shape[0] < inf_matrix.A.shape[1]:
        optimal_intensity = sol['inf_matrix'].fluence_1d_to_2d(fluence_1d=sol['optimal_intensity'])
        new_sol['optimal_intensity'] = inf_matrix.fluence_2d_to_1d(optimal_intensity)
    elif sol['optimal_intensity'].shape[0] == inf_matrix.A.shape[1]:
        new_sol['optimal_intensity'] = sol['optimal_intensity']
    else:
        raise ValueError("Beamlet resolution should be greater than or equal to beamlets for inf_matrix")

    # new_sol['dose_1d'] = inf_matrix.A * new_sol['optimal_intensity']

    # dose_3d = sol['inf_matrix'].dose_1d_to_3d(dose_1d=sol['dose_1d'])
    # new_sol['dose_1d'] = inf_matrix.dose_3d_to_1d(dose_3d=dose_3d)

    new_sol['inf_matrix'] = inf_matrix
    return new_sol





import numpy as np
from typing import List, Union


class Beams:
    """
    A class representing beams_dict.

    - **Attributes** ::

        :param beams_dict: beams_dict dictionary that contains information about the beams_dict in the format of
        dict: {
                   'ID': list(int),
                   'gantry_angle': list(float),
                   'collimator_angle': list(float) }
                  }

        :type beams_dict: dict

    - **Methods** ::

        :get_gantry_angle(beam_id: Optional(int, List[int]):
            Get gantry angle in degrees
        :get_collimator_angle(beam_id):
            Get collimator angle in degrees

    """

    def __init__(self, beams_dict):
        """

        :param beams_dict: Beams dictionary containing information about beams
        """

        self.beams_dict = beams_dict
        self.preprocess_beams()

    def get_beamlet_idx_2dgrid(self, beam_id: int) -> np.ndarray:
        """

        :param beam_id: beam_id for the beam
        :return: 2d grid of beamlets in 2.5*2.5 resolution
        """
        ind = self.beams_dict['ID'].index(beam_id)
        return self.beams_dict['beamlet_idx_2dgrid'][ind]

    def get_gantry_angle(self, beam_id: Union[int, List[int]]) -> Union[int, List[int]]:
        """
        Get gantry angle

        :param beam_id: beam_id for the beam
        :return: gantry angle for the beam_id
        """
        if isinstance(beam_id, int):
            ind = self.beams_dict['ID'].index(beam_id)
            return self.beams_dict['gantry_angle'][ind]
        elif isinstance(beam_id, list):
            return [self.beams_dict['gantry_angle'][self.beams_dict['ID'].index(idx)] for idx in beam_id]

    def get_collimator_angle(self, beam_id: Union[int, List[int]]) -> Union[int, List[int]]:
        """
        Get collimator angle

        :param beam_id: beam_id for the beam
        :return: collimator angle for the beam_id
        """
        if isinstance(beam_id, int):
            ind = self.beams_dict['ID'].index(beam_id)
            return self.beams_dict['collimator_angle'][ind]
        elif isinstance(beam_id, list):
            return [self.beams_dict['collimator_angle'][self.beams_dict['ID'].index(idx)] for idx in beam_id]

    @staticmethod
    def sort_beamlets(b_map):
        c = b_map[b_map >= 0]
        ind = np.arange(0, len(c))
        c_sort = np.sort(c)
        matrix_ind = [np.where(b_map == c_i) for c_i in c_sort]
        map_copy = b_map.copy()
        for i in range(len(ind)):
            map_copy[matrix_ind[i]] = ind[i]
        return map_copy

    def preprocess_beams(self):
        for i, beam_id in enumerate(self.beams_dict['ID']):
            ind = self.beams_dict['ID'].index(beam_id)
            beam_2d_grid = self.create_beamlet_idx_2d_grid(beam_id=beam_id)
            self.beams_dict.setdefault('beamlet_idx_2dgrid', []).append(beam_2d_grid)

    @staticmethod
    def get_original_map(beam_map):
        rowsNoRepeat = [0]
        for i in range(1, np.size(beam_map, 0)):
            if (beam_map[i, :] != beam_map[rowsNoRepeat[-1], :]).any():
                rowsNoRepeat.append(i)
        colsNoRepeat = [0]
        for j in range(1, np.size(beam_map, 1)):
            if (beam_map[:, j] != beam_map[:, colsNoRepeat[-1]]).any():
                colsNoRepeat.append(j)
        beam_map = beam_map[np.ix_(np.asarray(rowsNoRepeat), np.asarray(colsNoRepeat))]
        return beam_map

    def get_beamlet_width(self) -> float:
        """
        
        :return: beamlet width in the original beam
        """
        beamlets = self.beams_dict['beamlets'][0]
        return beamlets['width_mm'][0][0][0]

    def get_beamlet_height(self) -> float:
        """
        
        :return: beamlet height in the original beam
        """
        beamlets = self.beams_dict['beamlets'][0]
        return beamlets['height_mm'][0][0][0]

    def get_all_beam_ids(self) -> List[int]:
        return self.beams_dict['ID']

    def create_beamlet_idx_2d_grid(self, beam_id: int) -> np.ndarray:
        """
        Create 2d grid for the beamlets where each element is 2.5mm*2.5mm for the given beam id from x and y coordinates of beamlets.

        :param beam_id: beam_id for the beam
        :return: 2d grid of beamlets for the beam

        """
        ind = self.beams_dict['ID'].index(beam_id)
        beamlets = self.beams_dict['beamlets'][ind]
        x_positions = beamlets['position_x_mm'][0] - beamlets['width_mm'][0]/2  # x position is center of beamlet. Get left corner
        y_positions = beamlets['position_y_mm'][0] + beamlets['height_mm'][0]/2  # y position is center of beamlet. Get top corner
        right_ind = np.argmax(x_positions)
        bottom_ind = np.argmin(y_positions)
        w_all = np.column_stack((x_positions, y_positions))  # top left corners of all beamlets
        x_coord = np.arange(np.min(x_positions), np.max(x_positions) + beamlets['width_mm'][0][right_ind], 2.5)
        y_coord = np.arange(np.max(y_positions), np.min(y_positions) - beamlets['height_mm'][0][bottom_ind], -2.5)

        # create mesh grid in 2.5 mm resolution
        XX, YY = np.meshgrid(x_coord, y_coord)
        beamlet_idx_2d_grid = np.ones_like(XX, dtype=int)
        beamlet_idx_2d_grid = beamlet_idx_2d_grid * int(-1)  # make all elements to -1

        for row in range(XX.shape[0]):
            for col in range(XX.shape[1]):
                ind = np.where((w_all[:, 0] == XX[row, col]) & (w_all[:, 1] == YY[row, col]))  # find the position in matrix where we find the beamlet
                if np.size(ind) > 0:
                    ind = ind[0][0]
                    num_width = int(beamlets['width_mm'][0][ind]/2.5)
                    num_height = int(beamlets['height_mm'][0][ind] / 2.5)
                    beamlet_idx_2d_grid[row:row+num_height, col:col+num_width] = ind
        return beamlet_idx_2d_grid



from __future__ import annotations
import numpy as np
import cvxpy as cp
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from portpy.photon.plan import Plan
    from portpy.photon.influence_matrix import InfluenceMatrix


class CvxPyProb(object):
    """
    Optimization class for creating cvxpy problem object
    """

    def __init__(self, my_plan: Plan, inf_matrix: InfluenceMatrix = None, max_constraints: List[dict] = None,
                 mean_constraints: List[dict] = None, **opt_params):
        """
        It runs optimization to create optimal plan based upon clinical criteria

        :param my_plan: object of class Plan
        :param inf_matrix: object of class InfluenceMatrix
        :param solver: default solver 'MOSEK'. check cvxpy website for available solvers
        :param verbose: Default to True. If set to False, it will not print the solver iterations.
        :param cvxpy_options: cvxpy and the solver settings
        :param opt_params: optimization parameters for modifying parameters of problem statement
        :return: cvxpy problem object

        """

        # get data for optimization
        if max_constraints is None:
            max_constraints = my_plan.clinical_criteria.get_criteria(
                name='max_dose')  # returns all the max dose criteria as a list
        if mean_constraints is None:
            mean_constraints = my_plan.clinical_criteria.get_criteria(
                name='mean_dose')  # returns all the mean dose criteria as a list

        if inf_matrix is None:
            inf_matrix = my_plan.inf_matrix

        A = inf_matrix.A
        num_fractions = my_plan.get_num_of_fractions()
        st = inf_matrix
        self.my_plan = my_plan
        self.inf_matrix = inf_matrix

        # Opt params
        ptv_overdose_weight = opt_params['ptv_overdose_weight'] if 'ptv_overdose_weight' in opt_params else 10000
        ptv_underdose_weight = opt_params['ptv_underdose_weight'] if 'ptv_underdose_weight' in opt_params else 100000
        smoothness_weight = opt_params['smoothness_weight'] if 'smoothness_weight' in opt_params else 10
        total_oar_weight = opt_params['total_oar_weight'] if 'total_oar_weight' in opt_params else 10
        all_vox = np.arange(A.shape[0])
        oar_voxels = all_vox[~np.isin(np.arange(A.shape[0]), st.get_opt_voxels_idx('PTV'))]
        oar_weights = np.ones(A[oar_voxels, :].shape[0])
        pres_per_frac = my_plan.get_prescription() / my_plan.get_num_of_fractions()
        [Qx, Qy, num_rows, num_cols] = self.get_smoothness_matrix(inf_matrix.beamlets_dict)
        smoothness_X_weight = 0.6
        smoothness_Y_weight = 0.4

        # Construct optimization problem
        obj = []
        constraints = []

        # Construct the problem.

        x = cp.Variable(A.shape[1], pos=True, name='x')
        self.x = x
        dO = cp.Variable(len(st.get_opt_voxels_idx('PTV')), pos=True)
        dU = cp.Variable(len(st.get_opt_voxels_idx('PTV')), pos=True)

        print('Objective Start')
        # Add target objective
        obj += [(1 / len(st.get_opt_voxels_idx('PTV'))) * (
                ptv_overdose_weight * cp.sum_squares(dO) + ptv_underdose_weight * cp.sum_squares(dU))]
        # Add smoothness objective
        obj += [smoothness_weight * (smoothness_X_weight * (1 / num_cols) * cp.sum_squares(Qx @ x) +
                                     smoothness_Y_weight * (1 / num_rows) * cp.sum_squares(Qy @ x))]
        # Add quadratic oar objective
        obj += [total_oar_weight * (1 / A[oar_voxels, :].shape[0]) * cp.sum_squares(
            cp.multiply(cp.sqrt(oar_weights), A[oar_voxels, :] @ x))]

        print('Objective done')

        print('Constraints Start')

        # Adding max constraints
        for i in range(len(max_constraints)):
            if 'max_dose' in max_constraints[i]['name']:
                if 'limit_dose_gy' in max_constraints[i]['constraints']:
                    limit = max_constraints[i]['constraints']['limit_dose_gy']
                    org = max_constraints[i]['parameters']['structure_name']
                    if org != 'GTV' or org != 'CTV':
                        if org in my_plan.structures.structures_dict['name']:
                            constraints += [A[st.get_opt_voxels_idx(org), :] @ x <= limit / num_fractions]
                        else:
                            print('Structure {} not available!'.format(org))

        # Adding mean constraints
        for i in range(len(mean_constraints)):
            if 'mean_dose' in mean_constraints[i]['name']:
                if 'limit_dose_gy' in mean_constraints[i]['constraints']:
                    limit = mean_constraints[i]['constraints']['limit_dose_gy']
                    org = mean_constraints[i]['parameters']['structure_name']
                    # mean constraints using voxel weights
                    if org in my_plan.structures.structures_dict['name']:
                        constraints += [(1 / sum(st.get_opt_voxels_size(org))) *
                                        (cp.sum((cp.multiply(st.get_opt_voxels_size(org), A[st.get_opt_voxels_idx(org),
                                                                                          :] @ x)))) <= limit / num_fractions]

                    else:
                        print('Structure {} not available!'.format(org))

        # Step 1 and 2 constraints
        constraints += [A[st.get_opt_voxels_idx('PTV'), :] @ x <= pres_per_frac + dO]
        constraints += [A[st.get_opt_voxels_idx('PTV'), :] @ x >= pres_per_frac - dU]
        self.obj = obj
        self.constraints = constraints

        print('Constraints done')

    def add_max(self, struct: str, dose_gy: float):
        """
        Add max constraints to the problem
        :param struct: structure name
        :param dose_gy: dose in Gy per fraction.
        :return:
        """
        A = self.inf_matrix.A
        st = self.inf_matrix
        x = self.x

        max_constraint = [A[st.get_opt_voxels_idx(struct), :] @ x <= dose_gy]
        self.add_constraints(max_constraint)

    def add_mean(self, struct: str, dose_gy: float):
        """
        Add mean constraints to the problem
        :param struct: structure name
        :param dose_gy: dose in Gy per fraction.
        :return:
        """
        A = self.inf_matrix.A
        st = self.inf_matrix
        x = self.x

        mean_constraint = [(1 / sum(st.get_opt_voxels_size(struct))) *
                           (cp.sum((cp.multiply(st.get_opt_voxels_size(struct),
                                                A[st.get_opt_voxels_idx(struct),
                                                :] @ x)))) <= dose_gy]
        self.add_constraints(mean_constraint)

    def add_overdose_quad(self, struct: str, dose_gy: float, weight: float = 10000):
        """
        Add quadratic loss for the overdose voxels of the structure
        :param struct: structure name
        :param dose_gy: dose in Gy per fraction.
        :param weight: penalty/weight in the objective for overdose
        :return:
        """
        A = self.inf_matrix.A
        st = self.inf_matrix
        x = self.x

        dO = cp.Variable(len(st.get_opt_voxels_idx(struct)), pos=True, name='{}_overdose'.format(struct))
        obj = (1 / len(st.get_opt_voxels_idx(struct))) * (weight * cp.sum_squares(dO))
        self.add_objective(obj)
        self.add_constraints([A[st.get_opt_voxels_idx('PTV'), :] @ x <= dose_gy + dO])

    def add_underdose_quad(self, struct: str, dose_gy: float, weight: float = 100000):
        """
        Add quadratic loss for the underdose voxels of the structure
        :param struct: structure name
        :param dose_gy: dose in Gy per fraction.
        :param weight: penalty/weight in the objective for underdose
        :return:
        """
        A = self.inf_matrix.A
        st = self.inf_matrix
        x = self.x

        dU = cp.Variable(len(st.get_opt_voxels_idx(struct)), pos=True, name='{}_underdose'.format(struct))
        obj = (1 / len(st.get_opt_voxels_idx(struct))) * (weight * cp.sum_squares(dU))
        self.add_objective(obj)
        self.add_constraints([A[st.get_opt_voxels_idx('PTV'), :] @ x >= dose_gy - dU])

    def add_quad(self, struct: str = None, voxels: np.ndarray = None, weight: float = 10,
                 voxels_weight: np.ndarray = None):

        """
        Add quadratic objective to the optimization problem
        :param struct: structure for which quadratic loss is added to objective function
        :param voxels: Default to None. If set, quadratic loss will be added for the given voxels
        :param weight: Default to 10. penalty in the objective function for the given structure.
        :param voxels_weight: weight for each voxel in the objective function
        :return:
        """
        A = self.inf_matrix.A
        st = self.inf_matrix
        x = self.x

        obj = 0
        if voxels is not None:
            if voxels_weight is None:
                raise Exception('Please input weight array for the input voxels')
            obj = (1 / A[voxels, :].shape[0]) * cp.sum_squares(
                cp.multiply(cp.sqrt(voxels_weight), A[voxels, :] @ x))
        if struct is not None:
            obj = (1 / len(st.get_opt_voxels_idx(struct))) * (
                    weight * cp.sum_squares(A[st.get_opt_voxels_idx(struct), :] @ x))
        self.add_objective(obj)

    def add_smoothness_quad(self, weight: int = 10, smoothness_X_weight: int = 0.6, smoothness_Y_weight: int = 0.4):
        """
        Add quadratic smoothness to the optimization problem
        :param weight: smoothness weight
        :param smoothness_X_weight: weight in X direction of MLC (parallel to MLC)
        :param smoothness_Y_weight: weight in Y direction of MLC (perpendicular to MLC)
        :return:
        """

        st = self.inf_matrix
        x = self.x

        [Qx, Qy, num_rows, num_cols] = self.get_smoothness_matrix(st.beamlets_dict)
        obj = weight * (
                smoothness_X_weight * (1 / num_cols) * cp.sum_squares(Qx @ x) + smoothness_Y_weight * (1 / num_rows)
                * cp.sum_squares(Qy @ x))
        self.add_objective(obj)

    def add_constraints(self, constraints: list):
        """
        Add constraint to the problem
        :param constraints: list of constraints
        :return:
        """
        self.constraints += constraints

    def add_objective(self, obj):
        """
        Add objective function to the problem
        :param obj: objective function expression using cvxpy
        :return:
        """
        if not isinstance(obj, list):
            obj = [obj]

        self.obj += obj

    def add_boo(self, num_beams: int):
        """
        Select optimal beams from set of beams using MIP
        :param num_beams: number of beams to be selected
        :return:
        """
        st = self.inf_matrix
        x = self.x

        #  Constraints for selecting beams
        # binary variable for selecting beams
        b = cp.Variable(len(st.beamlets_dict), boolean=True)

        constraints = []
        constraints += [cp.sum(b) <= num_beams]
        for i in range(len(st.beamlets_dict)):
            start_beamlet = st.beamlets_dict[i]['start_beamlet']
            end_beamlet = st.beamlets_dict[i]['end_beamlet']
            M = 50  # upper bound on the beamlet intensity
            constraints += [x[start_beamlet:end_beamlet] <= b[i] * M]

        self.add_constraints(constraints)

    def solve(self, *args, **kwargs):
        problem = cp.Problem(cp.Minimize(sum(self.obj)), constraints=self.constraints)
        problem.solve(*args, **kwargs)

    def get_sol(self) -> dict:
        """
        Return optimal solution and influence matrix associated with it in the form of dictionary
        :return:
        """
        return {'optimal_intensity': self.x.value, 'inf_matrix': self.inf_matrix}

    def add_dvh(self, dvh_constraint: list):

        A = self.inf_matrix.A
        st = self.inf_matrix
        x = self.x

        import pandas as pd
        df_dvh_criteria = pd.DataFrame()
        count = 0
        criteria = self.my_plan.clinical_criteria.clinical_criteria_dict['criteria']
        for i in range(len(dvh_constraint)):
            if 'dose_volume' in dvh_constraint[i]['name']:
                limit_key = self.matching_keys(dvh_constraint[i]['constraints'], 'limit')
                if limit_key in dvh_constraint[i]['constraints']:
                    df_dvh_criteria.at[count, 'structure'] = dvh_constraint[i]['parameters']['structure_name']
                    df_dvh_criteria.at[count, 'dose_gy'] = dvh_constraint[i]['parameters']['dose_gy']

                    # getting max dose_1d for the same structure
                    max_dose_struct = 1000
                    for j in range(len(criteria)):
                        if 'max_dose' in criteria[j]['name']:
                            if 'limit_dose_gy' in criteria[j]['constraints']:
                                org = criteria[j]['parameters']['structure_name']
                                if org == dvh_constraint[i]['parameters']['structure_name']:
                                    max_dose_struct = criteria[j]['constraints']['limit_dose_gy']
                    df_dvh_criteria.at[count, 'M'] = max_dose_struct - dvh_constraint[i]['parameters']['dose_gy']
                    if 'perc' in limit_key:
                        df_dvh_criteria.at[count, 'vol_perc'] = dvh_constraint[i]['constraints'][limit_key]
                    count = count + 1

        # binary variable for dvh constraints
        b_dvh = cp.Variable(
            len(np.concatenate([st.get_opt_voxels_idx(org) for org in df_dvh_criteria.structure.to_list()])),
            boolean=True)

        start = 0
        constraints = []
        for i in range(len(df_dvh_criteria)):
            struct, limit, v, M = df_dvh_criteria.loc[i, 'structure'], df_dvh_criteria.loc[i, 'dose_gy'], \
                                  df_dvh_criteria.loc[i, 'vol_perc'], df_dvh_criteria.loc[i, 'M']
            end = start + len(st.get_opt_voxels_idx(struct))
            frac = self.my_plan.structures.get_fraction_of_vol_in_calc_box(struct)
            constraints += [
                A[st.get_opt_voxels_idx(struct), :] @ x <= limit / self.my_plan.get_num_of_fractions()
                + b_dvh[start:end] * M / self.my_plan.get_num_of_fractions()]
            constraints += [b_dvh @ st.get_opt_voxels_size(struct) <= (v / frac) / 100 * sum(
                st.get_opt_voxels_size(struct))]
            start = end
        self.add_constraints(constraints=constraints)

    @staticmethod
    def get_smoothness_matrix(beamReq: List[dict]) -> (np.ndarray, np.ndarray, int, int):
        """
        Create smoothness matrix so that adjacent beamlets are smooth out to reduce MU

        :param beamReq: beamlets dictionary from the object of influence matrix class
        :returns: tuple(Qx, Qy) where
            Qx: first matrix have values 1 and -1 for neighbouring beamlets in X direction
            Qy: second matrix with values 1 and -1 for neighbouring beamlets in Y direction

        :Example:
        Qx = [[1 -1 0 0 0 0]
              [0 0 1 -1 0 0]
              [0 0 0 0 1 -1]]

        """
        sRow = np.zeros((beamReq[-1]['end_beamlet'] + 1, beamReq[-1]['end_beamlet'] + 1), dtype=int)
        sCol = np.zeros((beamReq[-1]['end_beamlet'] + 1, beamReq[-1]['end_beamlet'] + 1), dtype=int)
        num_rows = 0
        num_cols = 0
        for b in range(len(beamReq)):
            beam_map = beamReq[b]['beamlet_idx_2dgrid']

            rowsNoRepeat = [0]
            for i in range(1, np.size(beam_map, 0)):
                if (beam_map[i, :] != beam_map[rowsNoRepeat[-1], :]).any():
                    rowsNoRepeat.append(i)
            colsNoRepeat = [0]
            for j in range(1, np.size(beam_map, 1)):
                if (beam_map[:, j] != beam_map[:, colsNoRepeat[-1]]).any():
                    colsNoRepeat.append(j)
            beam_map = beam_map[np.ix_(np.asarray(rowsNoRepeat), np.asarray(colsNoRepeat))]
            num_rows = num_rows + beam_map.shape[0]
            num_cols = num_cols + beam_map.shape[1]
            for r in range(np.size(beam_map, 0)):
                startCol = 0
                endCol = np.size(beam_map, 1) - 2
                while (beam_map[r, startCol] == -1) and (startCol <= endCol):
                    startCol = startCol + 1
                while (beam_map[r, endCol] == -1) and (startCol <= endCol):
                    endCol = endCol - 1

                for c in range(startCol, endCol + 1):
                    ind = beam_map[r, c]
                    RN = beam_map[r, c + 1]
                    if ind * RN >= 0:
                        sRow[ind, ind] = int(1)
                        sRow[ind, RN] = int(-1)

            for c in range(np.size(beam_map, 1)):
                startRow = 0
                endRow = np.size(beam_map, 0) - 2
                while (beam_map[startRow, c] == -1) and (startRow <= endRow):
                    startRow = startRow + 1
                while (beam_map[endRow, c] == -1) and (startRow <= endRow):
                    endRow = endRow - 1
                for r in range(startRow, endRow + 1):
                    ind = beam_map[r, c]
                    DN = beam_map[r + 1, c]
                    if ind * DN >= 0:
                        sCol[ind, ind] = int(1)
                        sCol[ind, DN] = int(-1)
        return sRow, sCol, num_rows, num_cols

    @staticmethod
    def matching_keys(dictionary, search_string):
        get_key = None
        for key, val in dictionary.items():
            if search_string in key:
                get_key = key
        if get_key is not None:
            return get_key
        else:
            return ''



from __future__ import annotations
import numpy as np
import cvxpy as cp
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from portpy.photon.plan import Plan
    from portpy.photon.influence_matrix import InfluenceMatrix
import json


class CvxPyProb(object):
    """
    Optimization class for creating cvxpy problem object
    """

    def __init__(self, my_plan: Plan, inf_matrix: InfluenceMatrix = None, max_constraints: List[dict] = None,
                 mean_constraints: List[dict] = None, obj_from_file: str = None, **opt_params):
        """
        It runs optimization to create optimal plan based upon clinical criteria

        :param my_plan: object of class Plan
        :param inf_matrix: object of class InfluenceMatrix
        :param solver: default solver 'MOSEK'. check cvxpy website for available solvers
        :param verbose: Default to True. If set to False, it will not print the solver iterations.
        :param cvxpy_options: cvxpy and the solver settings
        :param opt_params: optimization parameters for modifying parameters of problem statement
        :return: cvxpy problem object

        """

        # get data for optimization
        if max_constraints is None:
            max_constraints = []
        if mean_constraints is None:
            mean_constraints = []

        if inf_matrix is None:
            inf_matrix = my_plan.inf_matrix

        A = inf_matrix.A
        num_fractions = my_plan.get_num_of_fractions()
        st = inf_matrix

        self.my_plan = my_plan
        self.inf_matrix = inf_matrix

        # create and add rind constraints
        rinds = ['RIND_0', 'RIND_1', 'RIND_2', 'RIND_3', 'RIND_4']
        if rinds[0] not in my_plan.structures.structures_dict[
            'name']:  # check if rind is already created. If yes, skip rind creation
            self.create_rinds(my_plan, size_mm=[5, 5, 20, 30, 500])
            self.set_rinds_opt_voxel_idx(my_plan,
                                         inf_matrix=inf_matrix)  # rind_0 is 5mm after PTV, rind_2 is 5 mm after rind_1, and so on..
        else:
            self.set_rinds_opt_voxel_idx(my_plan, inf_matrix=inf_matrix)

        rind_max_dose_perc = [1.1, 1.05, 0.9, 0.85, 0.75]
        for i, rind in enumerate(rinds):  # Add rind constraints
            parameters = {'structure_name': rind}
            total_pres = my_plan.get_prescription()
            constraints = {'limit_dose_gy': total_pres * rind_max_dose_perc[i]}
            my_plan.clinical_criteria.add_criterion(criterion='max_dose', parameters=parameters,
                                                    constraints=constraints)
            if len(max_constraints) > 0:  # update max constraints by adding rind max constraints
                max_constraints = my_plan.clinical_criteria.get_criteria(name='max_dose')

        # Opt params
        # # setting weights for oar objectives
        # all_vox = np.arange(A.shape[0])
        # oar_voxels = all_vox[~np.isin(np.arange(A.shape[0]), st.get_opt_voxels_idx('PTV'))]
        # oar_weights = np.ones(A[oar_voxels, :].shape[0])
        # if my_plan.get_disease_site() == 'Prostate':
        #     oar_weights[np.where(np.isin(oar_voxels, st.get_opt_voxels_idx('RECT_WALL')))] = 20
        #     oar_weights[np.where(np.isin(oar_voxels, st.get_opt_voxels_idx('BLAD_WALL')))] = 5
        #     oar_weights[np.where(np.isin(oar_voxels, st.get_opt_voxels_idx('RIND_0')))] = 3
        #     oar_weights[np.where(np.isin(oar_voxels, st.get_opt_voxels_idx('RIND_1')))] = 3
        # elif my_plan.get_disease_site() == 'Lung':
        #     oar_weights[np.where(np.isin(oar_voxels, st.get_opt_voxels_idx('CORD')))] = 10
        #     oar_weights[np.where(np.isin(oar_voxels, st.get_opt_voxels_idx('ESOPHAGUS')))] = 20
        #     if 'HEART' in st.opt_voxels_dict['name']:
        #         oar_weights[np.where(np.isin(oar_voxels, st.get_opt_voxels_idx('HEART')))] = 20
        #     oar_weights[np.where(np.isin(oar_voxels, st.get_opt_voxels_idx('RIND_0')))] = 3
        #     oar_weights[np.where(np.isin(oar_voxels, st.get_opt_voxels_idx('RIND_1')))] = 3
        # ptv_overdose_weight = opt_params['ptv_overdose_weight'] if 'ptv_overdose_weight' in opt_params else 10000
        # ptv_underdose_weight = opt_params['ptv_underdose_weight'] if 'ptv_underdose_weight' in opt_params else 100000
        # smoothness_weight = opt_params['smoothness_weight'] if 'smoothness_weight' in opt_params else 10
        # total_oar_weight = opt_params['total_oar_weight'] if 'total_oar_weight' in opt_params else 10

        # Construct optimization problem
        self.obj = []
        self.constraints = []

        # Construct the problem.
        if max_constraints is not None and mean_constraints is not None and obj_from_file is not None:
            self.x = cp.Variable(A.shape[1], pos=True, name='x')

        print('Objective Start')
        if obj_from_file is not None:
            f = open(obj_from_file)
            obj_params = json.load(f)
            for i in range(len(obj_params)):
                if obj_params[i]['type'] == 'overdose-quadratic':
                    struct = obj_params[i]['parameters']['structure_name']
                    pres_per_frac = obj_params[i]['parameters']['dose_gy']/my_plan.get_num_of_fractions()
                    weight = obj_params[i]['parameters']['weight']
                    if struct in my_plan.structures.structures_dict['name']:
                        self.add_overdose_quad(struct=struct, dose_gy=pres_per_frac, weight=weight)
                    else:
                        print('Structure {} not available!'.format(struct))
                elif obj_params[i]['type'] == 'underdose-quadratic':
                    struct = obj_params[i]['parameters']['structure_name']
                    pres_per_frac = obj_params[i]['parameters']['dose_gy'] / my_plan.get_num_of_fractions()
                    weight = obj_params[i]['parameters']['weight']
                    if struct in my_plan.structures.structures_dict['name']:
                        self.add_underdose_quad(struct=struct, dose_gy=pres_per_frac, weight=weight)
                    else:
                        print('Structure {} not available!'.format(struct))
                elif obj_params[i]['type'] == 'quadratic':
                    struct = obj_params[i]['parameters']['structure_name']
                    weight = obj_params[i]['parameters']['weight']
                    if struct == 'NON-TARGET':
                        all_vox = np.arange(A.shape[0])
                        oar_voxels = all_vox[~np.isin(np.arange(A.shape[0]), st.get_opt_voxels_idx('PTV'))]
                        oar_weights = np.ones(A[oar_voxels, :].shape[0])
                        self.add_quad(voxels=oar_voxels, voxels_weight=weight*oar_weights)
                    else:
                        if struct in my_plan.structures.structures_dict['name']:
                            self.add_quad(struct=struct, weight=weight)
                        else:
                            print('Structure {} not available!'.format(struct))
                elif obj_params[i]['type'] == 'smoothness-quadratic':
                    weight = obj_params[i]['parameters']['weight']
                    self.add_smoothness_quad(weight=weight)

        # self.add_overdose_quad(struct='PTV', dose_gy=pres_per_frac, weight=ptv_overdose_weight)
        # self.add_underdose_quad(struct='PTV', dose_gy=pres_per_frac, weight=ptv_underdose_weight)
        # self.add_quad(voxels=oar_voxels, voxels_weight=oar_weights * total_oar_weight)
        # self.add_smoothness_quad(weight=smoothness_weight)

        print('Objective done')
        print('Constraints Start')

        for i in range(len(max_constraints)):
            if 'max_dose' in max_constraints[i]['name']:
                if 'limit_dose_gy' in max_constraints[i]['constraints']:
                    limit = max_constraints[i]['constraints']['limit_dose_gy']
                    org = max_constraints[i]['parameters']['structure_name']
                    if org != 'GTV' or org != 'CTV':
                        if org in my_plan.structures.structures_dict['name']:
                            self.add_max(struct=org, dose_gy=(limit / num_fractions))
                        else:
                            print('Structure {} not available!'.format(org))

        for i in range(len(mean_constraints)):
            if 'mean_dose' in mean_constraints[i]['name']:
                if 'limit_dose_gy' in mean_constraints[i]['constraints']:
                    limit = mean_constraints[i]['constraints']['limit_dose_gy']
                    org = mean_constraints[i]['parameters']['structure_name']
                    # mean constraints using voxel weights
                    if org in my_plan.structures.structures_dict['name']:
                        self.add_mean(struct=org, dose_gy=(limit / num_fractions))
                    else:
                        print('Structure {} not available!'.format(org))
        print('Constraints done')

    def add_max(self, struct: str, dose_gy: float):
        """
        Add max constraints to the problem
        :param struct: structure name
        :param dose_gy: dose in Gy per fraction.
        :return:
        """
        A = self.inf_matrix.A
        st = self.inf_matrix
        x = self.x

        max_constraint = [A[st.get_opt_voxels_idx(struct), :] @ x <= dose_gy]
        self.add_constraints(max_constraint)

    def add_mean(self, struct: str, dose_gy: float):
        """
        Add mean constraints to the problem
        :param struct: structure name
        :param dose_gy: dose in Gy per fraction.
        :return:
        """
        A = self.inf_matrix.A
        st = self.inf_matrix
        x = self.x

        mean_constraint = [(1 / sum(st.get_opt_voxels_size(struct))) *
                           (cp.sum((cp.multiply(st.get_opt_voxels_size(struct),
                                                A[st.get_opt_voxels_idx(struct),
                                                :] @ x)))) <= dose_gy]
        self.add_constraints(mean_constraint)

    def add_overdose_quad(self, struct: str, dose_gy: float, weight: float = 10000):
        """
        Add quadratic loss for the overdose voxels of the structure
        :param struct: structure name
        :param dose_gy: dose in Gy per fraction.
        :param weight: penalty/weight in the objective for overdose
        :return:
        """
        A = self.inf_matrix.A
        st = self.inf_matrix
        x = self.x

        dO = cp.Variable(len(st.get_opt_voxels_idx(struct)), pos=True, name='{}_overdose'.format(struct))
        obj = (1 / len(st.get_opt_voxels_idx(struct))) * (weight * cp.sum_squares(dO))
        self.add_objective(obj)
        self.add_constraints([A[st.get_opt_voxels_idx('PTV'), :] @ x <= dose_gy + dO])

    def add_underdose_quad(self, struct: str, dose_gy: float, weight: float = 100000):
        """
        Add quadratic loss for the underdose voxels of the structure
        :param struct: structure name
        :param dose_gy: dose in Gy per fraction.
        :param weight: penalty/weight in the objective for underdose
        :return:
        """
        A = self.inf_matrix.A
        st = self.inf_matrix
        x = self.x

        dU = cp.Variable(len(st.get_opt_voxels_idx(struct)), pos=True, name='{}_underdose'.format(struct))
        obj = (1 / len(st.get_opt_voxels_idx(struct))) * (weight * cp.sum_squares(dU))
        self.add_objective(obj)
        self.add_constraints([A[st.get_opt_voxels_idx('PTV'), :] @ x >= dose_gy - dU])

    def add_quad(self, struct: str = None, voxels: np.ndarray = None, weight: float = 10,
                 voxels_weight: np.ndarray = None):

        """
        Add quadratic objective to the optimization problem
        :param struct: structure for which quadratic loss is added to objective function
        :param voxels: Default to None. If set, quadratic loss will be added for the given voxels
        :param weight: Default to 10. penalty in the objective function for the given structure.
        :param voxels_weight: weight for each voxel in the objective function
        :return:
        """
        A = self.inf_matrix.A
        st = self.inf_matrix
        x = self.x

        obj = 0
        if voxels is not None:
            if voxels_weight is None:
                raise Exception('Please input weight array for the input voxels')
            obj = (1 / A[voxels, :].shape[0]) * cp.sum_squares(
                cp.multiply(cp.sqrt(voxels_weight), A[voxels, :] @ x))
        if struct is not None:
            obj = (1 / len(st.get_opt_voxels_idx(struct))) * (
                    weight * cp.sum_squares(A[st.get_opt_voxels_idx(struct), :] @ x))
        self.add_objective(obj)

    def add_smoothness_quad(self, weight: int = 10, smoothness_X_weight: int = 0.6, smoothness_Y_weight: int = 0.4):
        """
        Add quadratic smoothness to the optimization problem
        :param weight: smoothness weight
        :param smoothness_X_weight: weight in X direction of MLC (parallel to MLC)
        :param smoothness_Y_weight: weight in Y direction of MLC (perpendicular to MLC)
        :return:
        """

        st = self.inf_matrix
        x = self.x

        [Qx, Qy, num_rows, num_cols] = self.get_smoothness_matrix(st.beamlets_dict)
        obj = weight * (
                smoothness_X_weight * (1 / num_cols) * cp.sum_squares(Qx @ x) + smoothness_Y_weight * (1 / num_rows)
                * cp.sum_squares(Qy @ x))
        self.add_objective(obj)

    def add_constraints(self, constraints: list):
        """
        Add constraint to the problem
        :param constraints: list of constraints
        :return:
        """
        self.constraints += constraints

    def add_objective(self, obj):
        """
        Add objective function to the problem
        :param obj: objective function expression using cvxpy
        :return:
        """
        if not isinstance(obj, list):
            obj = [obj]

        self.obj += obj

    def add_boo(self, num_beams: int):
        """
        Select optimal beams from set of beams using MIP
        :param num_beams: number of beams to be selected
        :return:
        """
        st = self.inf_matrix
        x = self.x

        #  Constraints for selecting beams
        # binary variable for selecting beams
        b = cp.Variable(len(st.beamlets_dict), boolean=True)

        constraints = []
        constraints += [cp.sum(b) <= num_beams]
        for i in range(len(st.beamlets_dict)):
            start_beamlet = st.beamlets_dict[i]['start_beamlet']
            end_beamlet = st.beamlets_dict[i]['end_beamlet']
            M = 50  # upper bound on the beamlet intensity
            constraints += [x[start_beamlet:end_beamlet] <= b[i] * M]

        self.add_constraints(constraints)

    def solve(self, *args, **kwargs):
        problem = cp.Problem(cp.Minimize(sum(self.obj)), constraints=self.constraints)
        problem.solve(*args, **kwargs)

    def get_sol(self) -> dict:
        """
        Return optimal solution and influence matrix associated with it in the form of dictionary
        :return:
        """
        return {'optimal_intensity': self.x.value, 'inf_matrix': self.inf_matrix}

    def add_dvh(self, dvh_constraint: list):

        A = self.inf_matrix.A
        st = self.inf_matrix
        x = self.x

        import pandas as pd
        df_dvh_criteria = pd.DataFrame()
        count = 0
        criteria = self.my_plan.clinical_criteria.clinical_criteria_dict['criteria']
        for i in range(len(dvh_constraint)):
            if 'dose_volume' in dvh_constraint[i]['name']:
                limit_key = self.matching_keys(dvh_constraint[i]['constraints'], 'limit')
                if limit_key in dvh_constraint[i]['constraints']:
                    df_dvh_criteria.at[count, 'structure'] = dvh_constraint[i]['parameters']['structure_name']
                    df_dvh_criteria.at[count, 'dose_gy'] = dvh_constraint[i]['parameters']['dose_gy']

                    # getting max dose_1d for the same structure
                    max_dose_struct = 1000
                    for j in range(len(criteria)):
                        if 'max_dose' in criteria[j]['name']:
                            if 'limit_dose_gy' in criteria[j]['constraints']:
                                org = criteria[j]['parameters']['structure_name']
                                if org == dvh_constraint[i]['parameters']['structure_name']:
                                    max_dose_struct = criteria[j]['constraints']['limit_dose_gy']
                    df_dvh_criteria.at[count, 'M'] = max_dose_struct - dvh_constraint[i]['parameters']['dose_gy']
                    if 'perc' in limit_key:
                        df_dvh_criteria.at[count, 'vol_perc'] = dvh_constraint[i]['constraints'][limit_key]
                    count = count + 1

        # binary variable for dvh constraints
        b_dvh = cp.Variable(
            len(np.concatenate([st.get_opt_voxels_idx(org) for org in df_dvh_criteria.structure.to_list()])),
            boolean=True)

        start = 0
        constraints = []
        for i in range(len(df_dvh_criteria)):
            struct, limit, v, M = df_dvh_criteria.loc[i, 'structure'], df_dvh_criteria.loc[i, 'dose_gy'], \
                                  df_dvh_criteria.loc[i, 'vol_perc'], df_dvh_criteria.loc[i, 'M']
            end = start + len(st.get_opt_voxels_idx(struct))
            frac = self.my_plan.structures.get_fraction_of_vol_in_calc_box(struct)
            constraints += [
                A[st.get_opt_voxels_idx(struct), :] @ x <= limit / self.my_plan.get_num_of_fractions()
                + b_dvh[start:end] * M / self.my_plan.get_num_of_fractions()]
            constraints += [b_dvh @ st.get_opt_voxels_size(struct) <= (v / frac) / 100 * sum(
                st.get_opt_voxels_size(struct))]
            start = end
        self.add_constraints(constraints=constraints)

    @staticmethod
    def get_smoothness_matrix(beamReq: List[dict]) -> (np.ndarray, np.ndarray, int, int):
        """
        Create smoothness matrix so that adjacent beamlets are smooth out to reduce MU

        :param beamReq: beamlets dictionary from the object of influence matrix class
        :returns: tuple(Qx, Qy) where
            Qx: first matrix have values 1 and -1 for neighbouring beamlets in X direction
            Qy: second matrix with values 1 and -1 for neighbouring beamlets in Y direction

        :Example:
        Qx = [[1 -1 0 0 0 0]
              [0 0 1 -1 0 0]
              [0 0 0 0 1 -1]]

        """
        sRow = np.zeros((beamReq[-1]['end_beamlet'] + 1, beamReq[-1]['end_beamlet'] + 1), dtype=int)
        sCol = np.zeros((beamReq[-1]['end_beamlet'] + 1, beamReq[-1]['end_beamlet'] + 1), dtype=int)
        num_rows = 0
        num_cols = 0
        for b in range(len(beamReq)):
            beam_map = beamReq[b]['beamlet_idx_2dgrid']

            rowsNoRepeat = [0]
            for i in range(1, np.size(beam_map, 0)):
                if (beam_map[i, :] != beam_map[rowsNoRepeat[-1], :]).any():
                    rowsNoRepeat.append(i)
            colsNoRepeat = [0]
            for j in range(1, np.size(beam_map, 1)):
                if (beam_map[:, j] != beam_map[:, colsNoRepeat[-1]]).any():
                    colsNoRepeat.append(j)
            beam_map = beam_map[np.ix_(np.asarray(rowsNoRepeat), np.asarray(colsNoRepeat))]
            num_rows = num_rows + beam_map.shape[0]
            num_cols = num_cols + beam_map.shape[1]
            for r in range(np.size(beam_map, 0)):
                startCol = 0
                endCol = np.size(beam_map, 1) - 2
                while (beam_map[r, startCol] == -1) and (startCol <= endCol):
                    startCol = startCol + 1
                while (beam_map[r, endCol] == -1) and (startCol <= endCol):
                    endCol = endCol - 1

                for c in range(startCol, endCol + 1):
                    ind = beam_map[r, c]
                    RN = beam_map[r, c + 1]
                    if ind * RN >= 0:
                        sRow[ind, ind] = int(1)
                        sRow[ind, RN] = int(-1)

            for c in range(np.size(beam_map, 1)):
                startRow = 0
                endRow = np.size(beam_map, 0) - 2
                while (beam_map[startRow, c] == -1) and (startRow <= endRow):
                    startRow = startRow + 1
                while (beam_map[endRow, c] == -1) and (startRow <= endRow):
                    endRow = endRow - 1
                for r in range(startRow, endRow + 1):
                    ind = beam_map[r, c]
                    DN = beam_map[r + 1, c]
                    if ind * DN >= 0:
                        sCol[ind, ind] = int(1)
                        sCol[ind, DN] = int(-1)
        return sRow, sCol, num_rows, num_cols

    @staticmethod
    def create_rinds(my_plan, inf_matrix=None, size_mm: list = None, base_structure='PTV'):
        """
        :param inf_matrix: object of class inf_matrix
        :param my_plan: object of class Plan
        :param size_mm: size in mm. default size = [5,5,20,30, inf]. It means rind_0 is 5mm after PTV, rind_2 is 5 mm after rind_1, and so on..
        :param base_structure: structure from which rinds should be created
        :return: save rinds to plan object
        """
        if size_mm is None:
            size_mm = [5, 5, 20, 30, 500]
        if inf_matrix is None:
            inf_matrix = my_plan.inf_matrix
        print('creating rinds of size {} mm ..'.format(size_mm))
        if isinstance(size_mm, np.ndarray):
            size_mm = list(size_mm)
        ct_to_dose_map = inf_matrix.opt_voxels_dict['ct_to_dose_voxel_map'][0]
        dose_mask = ct_to_dose_map >= 0
        dose_mask = dose_mask.astype(int)
        my_plan.structures.create_structure('dose_mask', dose_mask)

        for ind, s in enumerate(size_mm):
            rind_name = 'RIND_{}'.format(ind)
            dummy_name = 'dummy_{}'.format(ind)
            if ind == 0:
                my_plan.structures.expand(base_structure, margin_mm=s, new_structure=dummy_name)
                my_plan.structures.subtract(dummy_name, base_structure, str1_sub_str2=rind_name)
                # inf_matrix.set_opt_voxel_idx(my_plan, struct=rind_name)
            else:
                # prev_rind_name = 'RIND_{}'.format(ind - 1)
                prev_dummy_name = 'dummy_{}'.format(ind - 1)
                my_plan.structures.expand(prev_dummy_name, margin_mm=s, new_structure=dummy_name)
                my_plan.structures.subtract(dummy_name, prev_dummy_name, str1_sub_str2=rind_name)
                my_plan.structures.delete_structure(prev_dummy_name)
                # for the last delete dummy
                if ind == len(size_mm) - 1:
                    my_plan.structures.delete_structure(dummy_name)  # delete dummy structure for last rind
            my_plan.structures.intersect(rind_name, 'dose_mask', str1_and_str2=rind_name)

        my_plan.structures.delete_structure('dose_mask')
        # inf_matrix.set_opt_voxel_idx(my_plan, struct=rind_name)
        print('rinds created!!')

    @staticmethod
    def set_rinds_opt_voxel_idx(plan_obj: Plan, inf_matrix: InfluenceMatrix):
        """

        :param plan_obj: object of class plan
        :param inf_matrix: object of class influence matrix
        :return: set rinds index based upon the influence matrix
        """
        rinds = [rind for idx, rind in enumerate(plan_obj.structures.structures_dict['name']) if 'RIND' in rind]
        for rind in rinds:
            inf_matrix.set_opt_voxel_idx(plan_obj, structure_name=rind)

    @staticmethod
    def matching_keys(dictionary, search_string):
        get_key = None
        for key, val in dictionary.items():
            if search_string in key:
                get_key = key
        if get_key is not None:
            return get_key
        else:
            return ''



from __future__ import annotations
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
from skimage import measure
from tabulate import tabulate
from portpy.photon.evaluation import Evaluation
from matplotlib.lines import Line2D
import os
from portpy.photon.utils import load_metadata
from portpy.photon.utils import view_in_slicer_jupyter
import pandas as pd
import webbrowser
from pathlib import Path
from typing import List, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from portpy.photon.plan import Plan
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


# mpl.use('TkAgg')  # or can use 'Qt5Agg', whatever you have/prefer


class Visualization:
    dose_type = Literal["Absolute(Gy)", "Relative(%)"]
    volume_type = Literal["Absolute(cc)", "Relative(%)"]

    @staticmethod
    def plot_robust_dvh(my_plan, dose_list, structs=None, style='solid', norm_flag=False, norm_volume=90,
                        norm_struct='PTV', weight_flag=True, plot_scenario=None, width=None, colors=None,
                        figsize=(12, 8), legend_font_size=10, title=None, filename=None, show=False, *args, **kwargs):
        if not isinstance(dose_list, list):
            dose_list = [dose_list]
        if len(dose_list) == 0:
            raise ValueError("dose_list is empty")
        if width is None:
            if style == 'dotted' or style == 'dashed':
                width = 2.5
            else:
                width = 2
        if colors is None:
            colors = my_plan.get_colors(30)
        if structs is None:
            # orgs = []
            structs = my_plan.structures.structures_dict['name']
        max_dose = 0.0
        all_orgs = my_plan.structures.structures_dict['name']
        # orgs = [struct.upper for struct in orgs]
        pres = my_plan.clinical_criteria['pres_per_fraction_gy'] * my_plan.clinical_criteria[
            'num_of_fractions']
        legend = []
        fig = plt.figure(figsize=figsize)
        plt.rcParams['font.size'] = 12
        for i in range(np.size(all_orgs)):
            if all_orgs[i] not in structs:
                continue
            dose_sort_list = []
            for dose in dose_list:
                x, y = my_plan.get_dvh(dose, all_orgs[i], weight_flag=weight_flag)
                dose_sort_list.append(x)
            d_sort_mat = np.column_stack(dose_sort_list)
            # Compute min/max DVH curve taken across scenarios.
            d_min_mat = np.min(d_sort_mat, axis=1)
            d_max_mat = np.max(d_sort_mat, axis=1)

            # Plot user-specified scenarios.
            if plot_scenario is not None:
                if not isinstance(dose_list, list):
                    plot_scenario = [plot_scenario]

                for n in range(len(plot_scenario)):
                    scene_num = plot_scenario[n]
                    if norm_flag:
                        norm_factor = Evaluation.get_dose(dose_list[scene_num], my_plan, norm_struct,
                                                          norm_volume) / pres
                        dose_sort_list[scene_num] = dose_sort_list[scene_num] / norm_factor
                        d_min_mat = d_min_mat / norm_factor
                        d_max_mat = d_max_mat / norm_factor
                    plt.plot(dose_sort_list[scene_num], 100 * y, linestyle=style, color=colors[i], linewidth=width,
                             *args, **kwargs)
            max_dose = np.maximum(max_dose, d_max_mat[-1])
            plt.plot(d_min_mat, 100 * y, linestyle='dotted', linewidth=width, color=colors[i], *args, **kwargs)
            plt.plot(d_max_mat, 100 * y, linestyle='dotted', linewidth=width, color=colors[i], *args, **kwargs)
            plt.fill_betweenx(100 * y, d_min_mat, d_max_mat, alpha=0.25, label=all_orgs[i], color=colors[i], *args,
                              **kwargs)
            legend.append(all_orgs[i])

        plt.xlabel('Dose (Gy)')
        plt.ylabel('Volume Fraction (%)')
        plt.xlim(0, max_dose * 1.1)
        plt.ylim(0, 100)
        plt.legend(prop={'size': legend_font_size}, loc="upper right")
        plt.grid(visible=True, which='major', color='#666666', linestyle='-')

        # Show the minor grid lines with very faint and almost transparent grey lines
        plt.minorticks_on()
        plt.grid(visible=True, which='minor', color='#999999', linestyle='--', alpha=0.2)
        y = np.arange(0, 101)
        if norm_flag:
            x = my_plan.clinical_criteria['pres_per_fraction_gy'] * my_plan.clinical_criteria[
                'num_of_fractions'] * np.ones_like(y)
        else:
            x = my_plan.clinical_criteria['pres_per_fraction_gy'] * np.ones_like(y)
        plt.plot(x, y, color='black')
        if title:
            plt.title(title)
        if show:
            plt.show()
        if filename is not None:
            fig.savefig(filename, bbox_inches="tight", dpi=300)

    @staticmethod
    def plot_dvh(my_plan: Plan, sol: dict = None, dose_1d: np.ndarray = None, structs: List[str] = None,
                 dose_scale: dose_type = "Absolute(Gy)",
                 volume_scale: volume_type = "Relative(%)", **options):
        """
        Create dvh plot for the selected structures

        :param my_plan: object of class Plan
        :param sol: optimal sol dictionary
        :param dose_1d: dose_1d in 1d voxels
        :param structs: structures to be included in dvh plot
        :param volume_scale: volume scale on y-axis. Default= Absolute(cc). e.g. volume_scale = "Absolute(cc)" or volume_scale = "Relative(%)"
        :param dose_scale: dose_1d scale on x axis. Default= Absolute(Gy). e.g. dose_scale = "Absolute(Gy)" or dose_scale = "Relative(%)"
        :keyword style (str): line style for dvh curve. default "solid". can be "dotted", "dash-dotted".
        :keyword width (int): width of line. Default 2
        :keyword colors(list): list of colors
        :keyword legend_font_size: Set legend_font_size. default 10
        :keyword figsize: Set figure size for the plot. Default figure size (12,8)
        :keyword create_fig: Create a new figure. Default True. If False, append to the previous figure
        :keyword title: Title for the figure
        :keyword filename: Name of the file to save the figure in current directory
        :keyword show: Show the figure. Default is True. If false, next plot can be append to it
        :keyword norm_flag: Use to normalize the plan. Default is False.
        :keyword norm_volume: Use to set normalization volume. default is 90 percentile.
        :return: dvh plot for the selected structures

        :Example:
        >>> Visualization.plot_dvh(my_plan, sol=sol, structs=['PTV', 'ESOPHAGUS'], dose_scale='Absolute(Gy)',volume_scale="Relative(%)", show=False, create_fig=True )
        """

        if dose_1d is None:
            if 'dose_1d' not in sol:
                dose_1d = sol['inf_matrix'].A @ (sol['optimal_intensity'] * my_plan.get_num_of_fractions())
            else:
                dose_1d = sol['dose_1d']

        if sol is None:
            sol = dict()
            sol['inf_matrix'] = my_plan.inf_matrix  # create temporary solution

        # getting options_fig:
        style = options['style'] if 'style' in options else 'solid'
        width = options['width'] if 'width' in options else None
        colors = options['colors'] if 'colors' in options else None
        legend_font_size = options['legend_font_size'] if 'legend_font_size' in options else 10
        figsize = options['figsize'] if 'figsize' in options else (12, 8)
        title = options['title'] if 'title' in options else None
        filename = options['filename'] if 'filename' in options else None
        show = options['show'] if 'show' in options else False
        # create_fig = options['create_fig'] if 'create_fig' in options else False
        show_criteria = options['show_criteria'] if 'show_criteria' in options else None
        ax = options['ax'] if 'ax' in options else None

        # getting norm options
        norm_flag = options['norm_flag'] if 'norm_flag' in options else False
        norm_volume = options['norm_volume'] if 'norm_volume' in options else 90
        norm_struct = options['norm_struct'] if 'norm_struct' in options else 'PTV'

        plt.rcParams['font.size'] = 12
        if width is None:
            if style == 'dotted' or style == 'dashed':
                width = 2.5
            else:
                width = 2
        if colors is None:
            colors = Visualization.get_colors()
        if structs is None:
            # orgs = []
            structs = my_plan.structures.structures_dict['name']
        max_dose = 0.0
        max_vol = 0.0
        all_orgs = my_plan.structures.structures_dict['name']
        # orgs = [struct.upper for struct in orgs]
        pres = my_plan.get_prescription()
        legend = []

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        if norm_flag:
            norm_factor = Evaluation.get_dose(sol, dose_1d=dose_1d, struct=norm_struct, volume_per=norm_volume) / pres
            dose_1d = dose_1d / norm_factor
        for i in range(np.size(all_orgs)):
            if all_orgs[i] not in structs:
                continue
            # for dose_1d in dose_list:
            #
            x, y = Evaluation.get_dvh(sol, struct=all_orgs[i], dose_1d=dose_1d)
            if dose_scale == 'Absolute(Gy)':
                max_dose = np.maximum(max_dose, x[-1])
                ax.set_xlabel('Dose (Gy)')
            elif dose_scale == 'Relative(%)':
                x = x / pres * 100
                max_dose = np.maximum(max_dose, x[-1])
                ax.set_xlabel('Dose (%)')

            if volume_scale == 'Absolute(cc)':
                y = y * my_plan.structures.get_volume_cc(all_orgs[i]) / 100
                max_vol = np.maximum(max_vol, y[1] * 100)
                ax.set_ylabel('Volume (cc)')
            elif volume_scale == 'Relative(%)':
                max_vol = np.maximum(max_vol, y[0] * 100)
                ax.set_ylabel('Volume Fraction (%)')
            ax.plot(x, 100 * y, linestyle=style, linewidth=width, color=colors[i])
            legend.append(all_orgs[i])

        if show_criteria is not None:
            for s in range(len(show_criteria)):
                if 'dose_volume' in show_criteria[s]['name']:
                    x = show_criteria[s]['parameters']['dose_gy']
                    y = show_criteria[s]['constraints']['limit_volume_perc']
                    ax.plot(x, y, marker='x', color='red', markersize=20)
        # plt.xlabel('Dose (Gy)')
        # plt.ylabel('Volume Fraction (%)')
        ax.set_xlim(0, max_dose * 1.1)
        ax.set_ylim(0, max_vol)
        ax.legend(legend, prop={'size': legend_font_size}, loc="upper right")
        ax.grid(visible=True, which='major', color='#666666', linestyle='-')

        # Show the minor grid lines with very faint and almost transparent grey lines
        # plt.minorticks_on()
        ax.minorticks_on()
        plt.grid(visible=True, which='minor', color='#999999', linestyle='--', alpha=0.2)
        y = np.arange(0, 101)
        # if norm_flag:
        #     x = pres * np.ones_like(y)
        # else:
        if dose_scale == "Absolute(Gy)":
            x = pres * np.ones_like(y)
        else:
            x = 100 * np.ones_like(y)

        ax.plot(x, y, color='black')
        if title:
            ax.set_title(title)
        if show:
            plt.show()
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight", dpi=300)
        return ax

    @staticmethod
    def plot_binary_mask_points(my_plan, structure: str, show: bool = True, color: List[str] = None):
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        ind = my_plan.structures.structures_dict['name'].index(structure)
        mask_3d = my_plan.structures.structures_dict['structure_mask_3d'][ind]
        pos = np.where(mask_3d == 1)
        if color is None:
            color = Visualization.get_colors()
        ax.scatter(pos[0], pos[1], pos[2], c=color)
        if show:
            plt.show()

    @staticmethod
    def get_colors():
        """

        :return: return list of 20 colors
        """
        colors = ['#3cb44b', '#ffe119', '#4363d8', '#f58231', '#911eb4',
                  '#46f0f0', '#f032e6', '#bcf60c', '#fabebe', '#008080', '#e6beff',
                  '#9a6324', '#fffac8', '#800000', '#aaffc3', '#808000', '#ffd8b1',
                  '#000075', '#808080', '#ffffff', '#000000', '#e6194b']
        return colors

    @staticmethod
    def surface_plot(matrix: np.ndarray, **kwargs):
        # acquire the cartesian coordinate matrices from the matrix
        # x is cols, y is rows
        (x, y) = np.meshgrid(np.arange(matrix.shape[0]), np.arange(matrix.shape[1]))
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        surf = ax.plot_surface(x, y, np.transpose(matrix), **kwargs)
        return ax, surf

    @staticmethod
    def plot_fluence_2d(beam_id: int, sol: dict, **options):
        """

        Displays fluence in 2d for the given beam_id

        :param beam_id: beam_id of the beam
        :param sol: solution dictionary after optimization
        :return: 2d optimal fluence plot

        :Example:
        >>> Visualization.plot_fluence_2d(beam_id=0, sol=sol, **options)
        """
        return sol['inf_matrix'].plot_fluence_2d(beam_id=beam_id, sol=sol, **options)

    @staticmethod
    def plot_fluence_3d(beam_id: int, sol: dict, **options):
        """
        Displays fluence in 3d for the given beam_id

        :param sol: solution after optimization
        :param beam_id: beam_id of the beam
        :return: 3d optimal fluence plot

        :Example:
        >>> Visualization.plot_fluence_3d(beam_id=0, sol=sol, **options)
        """
        return sol['inf_matrix'].plot_fluence_3d(beam_id=beam_id, sol=sol, **options)

    @staticmethod
    def plot_2d_dose(my_plan: Plan, sol: dict, slice_num: int = 40, structs: List[str] = None, show_dose: bool = True,
                     show_struct: bool = True, show_isodose: bool = False,
                     **options) -> None:
        """

        Plot 2d view of ct, dose_1d, isodose and structure contours

        :param sol: solution to optimization
        :param my_plan: object of class Plan
        :param slice_num: slice number
        :param structs: structures for which contours to be displayed on the slice view. e.g. structs = ['PTV, ESOPHAGUS']
        :param show_dose: view dose_1d on the slice
        :param show_struct: view structure on the slice
        :param show_isodose: view isodose
        :param dpi: Default dpi=100 for figure
        :return: plot 2d view of ct, dose_1d, isodose and structure contours

        :Example:
        >>> Visualization.plot_2d_dose(my_plan, sol=sol, slice_num=50, structs=['PTV'], show_isodose=False)
        """

        # getting options_fig:
        figsize = options['figsize'] if 'figsize' in options else (8, 8)
        title = options['title'] if 'title' in options else None
        filename = options['filename'] if 'filename' in options else None
        dpi = options['dpi'] if 'dpi' in options else 100
        show = options['show'] if 'show' in options else False
        ax = options['ax'] if 'ax' in options else None

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
        plt.rcParams["figure.autolayout"] = True
        ct = my_plan.ct['ct_hu_3d'][0]

        # adjust the main plot to make room for the legends
        plt.subplots_adjust(left=0.2)
        dose_3d = []
        if show_dose:
            dose_1d = sol['inf_matrix'].A @ (sol['optimal_intensity']*my_plan.get_num_of_fractions())
            dose_3d = sol['inf_matrix'].dose_1d_to_3d(dose_1d=dose_1d)
            ax.imshow(ct[slice_num, :, :], cmap='gray')
            masked = np.ma.masked_where(dose_3d[slice_num, :, :] <= 0, dose_3d[slice_num, :, :])
            im = ax.imshow(masked, alpha=0.4, interpolation='none',
                           cmap='rainbow')

            plt.colorbar(im, ax=ax, pad=0.1)

        if show_isodose:
            if not show_dose:
                dose_1d = sol['inf_matrix'].A * sol['optimal_intensity'] * my_plan.get_num_of_fractions()
                dose_3d = sol['inf_matrix'].dose_1d_to_3d(dose_1d=dose_1d)
            dose_legend = Visualization.legend_dose_storage(my_plan)
            ax.contour(dose_3d[slice_num, :, :], dose_legend['dose_1d value'],
                       colors=dose_legend['dose_1d color'],
                       linewidths=0.5, zorder=2)
            dose_list = []
            for item in range(0, len(dose_legend['dose_1d name'])):
                dose_list.append(Line2D([0], [0],
                                        color=dose_legend['dose_1d color'][item],
                                        lw=1,
                                        label=dose_legend['dose_1d name'][item]))
            ax.add_artist(ax.legend(handles=dose_list,
                                    bbox_to_anchor=(1.15, 1), loc='upper left', borderaxespad=0.))
        if title is not None:
            ax.set_title('{}: Axial View - Slice #: {}'.format(title, slice_num))
        else:
            ax.set_title('Axial View - Slice #: {}'.format(slice_num))

        if show_struct:
            if structs is None:
                structs = my_plan.structures.structures_dict['name']
            struct_masks = my_plan.structures.structures_dict['structure_mask_3d']
            all_mask = []
            colors = Visualization.get_colors()
            for i in range(len(structs)):
                ind = my_plan.structures.structures_dict['name'].index(structs[i])
                cmap = mpl.colors.ListedColormap(colors[i])
                contours = measure.find_contours(struct_masks[ind][slice_num, :, :], 0.5)
                for contour in contours:
                    ax.plot(contour[:, 1], contour[:, 0], linewidth=2, color=colors[i])
            labels = [struct for struct in structs]
            # get the colors of the values, according to the
            import matplotlib.patches as mpatches
            patches = [mpatches.Patch(color=colors[i], label=labels[i]) for i in range(len(labels))]
            # rax.labels = labels
            ax.legend(handles=patches, bbox_to_anchor=(0.1, 0.8), loc=2, borderaxespad=0.)
                       # bbox_transform=fig.transFigure)
        if show:
            plt.show()
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight", dpi=300)
        return ax

    @staticmethod
    def get_cmap_colors(n, name='hsv'):
        """Returns a function that maps each index in 0, 1, ..., n-1 to a distinct
        RGB color; the keyword argument name must be a standard mpl colormap name."""
        return plt.cm.get_cmap(name, n)

    @staticmethod
    def legend_dose_storage(my_plan: Plan) -> dict:
        # dose_color = [[0.55, 0, 1], [0, 0, 1], [0, 0.5, 1], [0, 1, 0],
        #               [1, 1, 0], [1, 0.65, 0], [1, 0, 0], [0.55, 0, 0]]
        dose_color = [[0.55, 0, 0], [0, 0, 1], [0.55, 0, 1], [0, 0.5, 1], [0, 1, 0], [1, 0, 0]]
        dose_level = [0.3, 0.5, 0.7, 0.9, 1.0, 1.1]
        dose_prescription = my_plan.clinical_criteria.clinical_criteria_dict['pres_per_fraction_gy'] * \
                            my_plan.clinical_criteria.clinical_criteria_dict['num_of_fractions']
        dose_value = [item * dose_prescription for item in dose_level]
        dose_name = []
        for item in range(0, len(dose_level)):
            dose_name.append(str(round(dose_level[item] * 100, 2)) + ' % / ' +
                             str(round(dose_value[item], 3)) + ' ' + 'Gy')
        dose_storage_legend = {'dose_1d color': dose_color, 'dose_1d level': dose_level, 'dose_1d value': dose_value,
                               'dose_1d name': dose_name}
        return dose_storage_legend

    @staticmethod
    def view_in_slicer(my_plan: Plan, slicer_path: str = None, data_dir: str = None) -> None:
        """

        :param my_plan: object of class Plan
        :param slicer_path: slicer executable path on your local machine
        :param data_dir: the folder path where data located, defaults to None.
                If path = None, then it assumes the data is in sub-folder named data in the current directory
        :return: plot the images in 3d slicer

        view ct, dose_1d and structure set images in 3d slicer

        :Example:
        >>> Visualization.view_in_slicer(my_plan, slicer_path='C:/Slicer/Slicer.exe', data_dir='path/to/nrrd/image')
        """
        if slicer_path is None:
            slicer_path = r'C:\ProgramData\NA-MIC\Slicer 4.11.20210226\Slicer.exe'
        if data_dir is None:
            data_dir = os.path.join('../..', 'data', my_plan.patient_id)
        if not os.path.exists(data_dir):  # check if valid directory
            raise Exception("Invalid data directory. Please input valid directory")
        slicer_script_dir = os.path.join(Path(__file__).parents[0], 'utils', 'slicer_script.py')
        import subprocess
        subprocess.run([slicer_path, f"--python-script", slicer_script_dir, data_dir,
                        ','.join(my_plan.structures.structures_dict['name'])], shell=False)
        print('Done')

    @staticmethod
    def view_in_slicer_jupyter(my_plan: Plan, dose_1d: np.ndarray = None, sol: dict = None,
                               ct_name: str = 'ct', dose_name: str = 'dose', struct_set_name: str = 'rt_struct',
                               show_ct: bool = True,
                               show_dose: bool = True,
                               show_structs: bool = True):
        """
        This method helps to visualize CT, Dose and Rt struct in 3d slicer


        :param my_plan: object of class plan
        :param dose_1d: dose in 1d
        :param sol: solution dictionary
        :param ct_name: Default to 'ct'. name of the ct node in 3D slicer
        :param dose_name: Default to 'dose'. name of the dose node in 3D slicer
        :param struct_set_name: name of the rtstruct
        :param show_structs: default to True. If false, will not create structure node
        :param show_dose: default to True. If false, will not create dose node
        :param show_ct:default to True. If false, will not create ct node
        :return: visualize in slicer jupyter
        """
        view_in_slicer_jupyter.view_in_slicer_jupyter(my_plan, dose_1d=dose_1d, sol=sol, ct_name=ct_name, dose_name=dose_name,
                                                      struct_set_name=struct_set_name, show_ct=show_ct, show_dose=show_dose,
                                                      show_structs=show_structs)

    @staticmethod
    def display_patient_metadata(patient_id: str, data_dir: str = None,
                                 in_browser: bool = False, return_beams_df: bool = False,
                                 return_structs_df: bool = False):
        """Displays the patient information in console or html format. If in_browswer is enabled
        it creates a temporary html file and lnunches your browser

        :param patient_id: the patient id
        :param data_dir: the folder path where data located, defaults to None.
                If path = None, then it assumes the data is in sub-folder named data in the current directory
        :param in_browser: visualize in pretty way in browser. default to False. If false, plot table in console
        :param return_beams_df: return dataframe containing beams metadata
        :param return_structs_df: return dataframe containing structures metadata
        :raises invalid directory error: raises an exception if invalid data directory

        :Example:
        >>> Visualization.display_patient_metadata(patient_id=patient_id, data_dir='path/to/data', in_browser=True)
        """

        if data_dir is None:
            data_dir = os.path.join('../..', 'data')
            data_dir = os.path.join(data_dir, patient_id)
        else:
            data_dir = os.path.join(data_dir, patient_id)
        if not os.path.exists(data_dir):  # check if valid directory
            raise Exception("Invalid data directory. Please input valid directory")

        meta_data = load_metadata(data_dir)
        # if show_beams:
        # check if full and/or sparse influence matrices are provided.
        # Sparse matrix is just a truncated version of the full matrix (zeroing out small elements)
        # often used in the optimization for computational efficiency
        beams = meta_data['beams_dict']  # get beams_dict metadata
        beams_df = pd.DataFrame.from_dict(beams)  # using Pandas data structure
        is_full = beams_df['influenceMatrixFull_File'].str.contains('full',
                                                                    na=False)  # does the data include the full influence matrix
        is_sparse = beams_df['influenceMatrixSparse_File'].str.contains('sparse',
                                                                        na=False)  # does the data include the sparse influence matrix
        for ind, (sparse, full) in enumerate(zip(is_full, is_sparse)):  # create column for sparse/full check
            if sparse and full:
                beams_df.at[ind, 'influence_matrix(sparse/full)'] = 'Both'
            elif sparse and not full:
                beams_df.at[ind, 'influence_matrix(sparse/full)'] = 'Only Sparse'
            elif not sparse and full:
                beams_df.at[ind, 'influence_matrix(sparse/full)'] = 'Only Full'
        #  pick information to include in the table
        keep_columns = ['ID', 'gantry_angle', 'collimator_angle', 'couch_angle', 'beam_modality', 'energy_MV',
                        'influence_matrix(sparse/full)',
                        'iso_center', 'MLC_name',
                        'machine_name']
        beams_df = beams_df[keep_columns]

        structures = meta_data['structures']
        struct_df = pd.DataFrame.from_dict(structures)
        keep_columns = ['name', 'volume_cc']  # discard the columns except this columns
        struct_df = struct_df[keep_columns]

        if return_beams_df and return_structs_df:
            return beams_df, struct_df
        elif return_beams_df:
            return beams_df
        elif return_structs_df:
            return struct_df
        # Write the results in a temporary html file in the current directory and launch a browser to display
        if in_browser:
            style_file = os.path.join('../..', 'df_style.css')
            html_string = '''
                    <html>
                      <head><title>Portpy MetaData</title></head>
                      <link rel="stylesheet" type="text/css" href="{style}"/>
                      <body>
                      <h1> PortPy Metadata </h1> 
                      <h4> Beams Metadata </h4>
                        {table_1}
                      <h4> Structures Metadata </h4>
                        {table_2}
                      </body>
                    </html>.
                    '''  # create html body and append table to it
            with open('temp.html', 'w') as f:
                f.write(html_string.format(table_1=beams_df.to_html(index=False, header=True, classes='mystyle'),
                                           table_2=struct_df.to_html(index=False, header=True, classes='mystyle'),
                                           style=style_file))
            webbrowser.open('file://' + os.path.realpath('temp.html'))
        else:
            if Visualization.is_notebook():
                from IPython.display import display
                print('Beams table..')
                with pd.option_context('display.max_rows', None,
                                       'display.max_columns', None,
                                       'display.precision', 3,
                                       'display.colheader_justify', 'center',
                                       ):
                    beams_df = beams_df.style.set_properties(**{'text-align': 'center'})
                    display(beams_df)
                print('Structure table..')
                with pd.option_context('display.max_rows', None,
                                       'display.max_columns', None,
                                       'display.precision', 3,
                                       'display.colheader_justify', 'center',
                                       ):
                    struct_df = struct_df.style.set_properties(**{'text-align': 'center'})
                    display(struct_df)
            else:
                print('Beams table..')
                print(tabulate(beams_df, headers='keys', tablefmt='psql'))  # print the table in console using tabulate
                print('\n\nStructures table..')
                print(tabulate(struct_df, headers='keys', tablefmt='psql'))

    @staticmethod
    def display_patients(data_dir: str = None, in_browser: bool = False, return_df: bool = False):
        """
        Displays the list of patients included in data_dir folder

        :param data_dir: folder including patient data.
            If it is None, then it assumes the data is in the current directory under sub-folder named "data"
        :param in_browser: visualize in pretty way in browser. default to False. If false, plot table in console
        :param return_df: return dataframe instead of visualization
        :raises invalid directory error: raises an exception if invalid data directory.

        :return display patient information in table

        """

        display_dict = {}  # we add all the relevant information from meta_data to this dictionary
        if data_dir is None:  # if data directory not provided, then use the subfolder named "data" in the current directory
            data_dir = os.path.join('../..', 'data')
        if not os.path.exists(data_dir):  # check if valid directory
            raise Exception("Invalid data directory. Please input valid directory")
        pat_ids = os.listdir(data_dir)
        for i, pat_id in enumerate(pat_ids):  # Loop through patients in path
            if "Patient" in pat_id:  # ignore irrelevant folders
                display_dict.setdefault('patient_id', []).append(pat_id)
                meta_data = load_metadata(os.path.join(data_dir, pat_id))  # load metadata for the patients
                # set the keys and append to display dict
                display_dict.setdefault('disease_site', []).append(meta_data['clinical_criteria']['disease_site'])
                ind = meta_data['structures']['name'].index('PTV')
                display_dict.setdefault('ptv_vol_cc', []).append(meta_data['structures']['volume_cc'][ind])
                display_dict.setdefault('num_beams', []).append(len(meta_data['beams_dict']['ID']))
                # check if all the iso centers are same for beams_dict
                res = all(
                    ele == meta_data['beams_dict']['iso_center'][0] for ele in meta_data['beams_dict']['iso_center'])
                if res:
                    display_dict.setdefault('iso_center_shift ', []).append('No')
                else:
                    display_dict.setdefault('iso_center_shift ', []).append('Yes')
        df = pd.DataFrame.from_dict(display_dict)  # convert dictionary to dataframe
        if return_df:
            return df
        if in_browser:
            style_file = os.path.join('../..', 'df_style.css')  # get style file path
            html_string = '''
                    <html>
                      <head><title>Portpy MetaData</title></head>
                      <link rel="stylesheet" type="text/css" href="{style}"/>
                      <body>
                      <h4> Patients Metadata </h4>
                        {table}
                      </body>
                    </html>.
                    '''  # create html body and append table to it
            with open('temp.html', 'w') as f:
                f.write(
                    html_string.format(table=df.to_html(index=False, header=True, classes='mystyle'), style=style_file))
            webbrowser.open('file://' + os.path.realpath('temp.html'))
        else:
            if Visualization.is_notebook():
                from IPython.display import display
                with pd.option_context('display.max_rows', None,
                                       'display.max_columns', None,
                                       'display.precision', 3,
                                       ):

                    display(df)
            else:
                print(tabulate(df, headers='keys', tablefmt='psql'))  # print in console using tabulate

    @staticmethod
    def plan_metrics(my_plan: Plan, sol: Union[dict, List[dict]], html_file_name='temp.html', sol_names: List[str] = None,
                     return_df: bool = False):
        """
        Visualize the plan metrics for clinical criteria in browser.
        It evaluate the plan by comparing the metrics against required criteria.

        If plan value is green color. It meets all the Limits and Goals
        If plan value is yellow color. It meets limits but not goals
        If plan value is red color. It violates both limit and goals
 
        :param my_plan: object of class Plan
        :param sol: optimal solution dictionary
        :param html_file_name:  name of the html file to be launched in browser
        :param sol_names: Default to Plan Value. column names for the plan evaluation
        :param return_df: return df instead of visualization
        :return: plan metrics in browser
        """

        # convert clinical criteria in dataframe

        df = pd.DataFrame.from_dict(my_plan.clinical_criteria.clinical_criteria_dict['criteria'])
        if isinstance(sol, dict):
            sol = [sol]
        if sol_names is None:
            if len(sol) > 1:
                sol_names = ['Plan Value ' + str(i) for i in range(len(sol))]
            else:
                sol_names = ['Plan Value']
        for p, s in enumerate(sol):
            dose_1d = s['inf_matrix'].A @ (s['optimal_intensity'] * my_plan.get_num_of_fractions())
            for ind in range(len(df)):  # Loop through the clinical criteria
                if df.name[ind] == 'max_dose':
                    struct = df.parameters[ind]['structure_name']
                    if struct in my_plan.structures.get_structures():
                        max_dose = Evaluation.get_max_dose(s, dose_1d=dose_1d, struct=struct)  # get max dose_1d
                        if 'limit_dose_gy' in df.constraints[ind] or 'goal_dose_gy' in df.constraints[ind]:
                            df.at[ind, sol_names[p]] = max_dose
                        elif 'limit_dose_perc' in df.constraints[ind] or 'goal_dose_perc' in df.constraints[ind]:
                            df.at[ind, sol_names[p]] = max_dose / (
                                    my_plan.get_prescription() * my_plan.get_num_of_fractions()) * 100
                if df.name[ind] == 'mean_dose':
                    struct = df.parameters[ind]['structure_name']
                    if struct in my_plan.structures.get_structures():
                        mean_dose = Evaluation.get_mean_dose(s, dose_1d=dose_1d, struct=struct)
                        df.at[ind, sol_names[p]] = mean_dose
                if df.name[ind] == 'dose_volume_V':
                    struct = df.parameters[ind]['structure_name']
                    if struct in my_plan.structures.get_structures():
                        if 'limit_volume_perc' in df.constraints[ind] or 'goal_volume_perc' in df.constraints[ind]:
                            dose = df.parameters[ind]['dose_gy']
                            volume = Evaluation.get_volume(s, dose_1d=dose_1d, struct=struct, dose_value_gy=dose)
                            df.at[ind, sol_names[p]] = volume
                        elif 'limit_volume_cc' in df.constraints[ind] or 'goal_volume_cc' in df.constraints[ind]:
                            dose = df.parameters[ind]['dose_gy']
                            volume = Evaluation.get_volume(s, dose_1d=dose_1d, struct=struct, dose_value_gy=dose)
                            vol_cc = my_plan.structures.get_volume_cc(structure_name=struct) * volume / 100
                            df.at[ind, sol_names[p]] = vol_cc
        pd.set_option("display.precision", 3)
        df.dropna(axis=0, inplace=True)  # remove structures which are not present
        df.reset_index(drop=True, inplace=True)  # reset the index

        def color_plan_value(row):

            highlight_red = 'background-color: red;'
            highlight_green = 'background-color: green;'
            highlight_orange = 'background-color: orange;'
            default = ''

            def matching_keys(dictionary, search_string):  # method to find the key of only part of the key match
                """

                :param dictionary: dictionary to search the string
                :param search_string: string to be searched in dictionary
                :return: full key if found search string found else empty string
                """
                get_key = None
                for key, val in dictionary.items():
                    if search_string in key:
                        get_key = key
                if get_key is not None:
                    return get_key
                else:
                    return ''

            limit_key = matching_keys(row['constraints'], 'limit')
            goal_key = matching_keys(row['constraints'], 'goal')

            row_color = len(row)*[default]  # default color for all rows initially
            # must return one string per cell in this row
            for i in range(len(row) - 1):
                if limit_key in row['constraints']:
                    if not (pd.isnull(row[i])):
                        if row[i] > row['constraints'][limit_key] + 0.0001:  # added epsilon to avoid minor differences
                            row_color[i] = highlight_red  # make plan value in red
                        else:
                            row_color[i] = highlight_green  # make plan value in red
                if goal_key in row['constraints']:
                    if not (pd.isnull(row[i])):
                        if row[i] > row['constraints'][goal_key] + 0.0001:
                            row_color[i] = highlight_orange  # make plan value in red
                        else:
                            row_color[i] = highlight_green  # make plan value in red
            return row_color
        sol_names.append('constraints')
        styled_df = df.style.apply(color_plan_value, subset=sol_names, axis=1)  # apply
        # color to dataframe using df.style method
        if return_df:
            return styled_df
        if Visualization.is_notebook():
            from IPython.display import display
            with pd.option_context('display.max_rows', None,
                                   'display.max_columns', None,
                                   'display.precision', 3,
                                   ):

                display(styled_df)
        else:
            html = styled_df.render()  # render to html
            html_string = '''
                            <html>
                              <head><title>Portpy Clinical Criteria Evaluation</title></head>
                              <style> 
                                table, th, td {{font-size:10pt; border:1px solid black; border-collapse:collapse; text-align:left;}}
                                th, td {{padding: 5px;}}
                              </style>
                              <body>
                              <h1> Clinical Criteria</h1>
                              <h4 style="color: green">Meets limit and goal</h4>
                              <h4 style="color: orange">Meets limit but not goal</h4>
                              <h4 style="color: red">Violate both limit and goal</h4>
                                {table}
                              </body>
                            </html>.
                            '''
            with open(html_file_name, 'w') as f:
                f.write(html_string.format(table=html))
            webbrowser.open('file://' + os.path.realpath(html_file_name))

    @staticmethod
    def is_notebook() -> bool:
        try:
            from IPython import get_ipython
            shell = get_ipython().__class__.__name__
            if shell == 'ZMQInteractiveShell':
                return True  # Jupyter notebook or qtconsole
            elif shell == 'TerminalInteractiveShell':
                return False  # Terminal running IPython
            else:
                if 'google.colab' in str(get_ipython()):
                    return True
                else:
                    return False  # Other type (?)
        except NameError:
            return False  # Probably standard Python interpreter
        except:
            return False  # Probably standard Python interpreter